# Utility functions for regularization methods
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


import copy
import numpy as np
import time
from scipy import signal
from scipy.interpolate import griddata
import warnings

import jax
import jax.numpy as jnp
from jax import lax, jit, vmap

from utax.wavelet import WaveletTransform
from herculens.Util import vkl_util



def interp_unstruct_grid(image, x, y, new_x, new_y):
    interpolated_image = griddata((x.flatten(), y.flatten()), 
                                  image.flatten(), 
                                  (new_x.flatten(), new_y.flatten()), 
                                  method='linear')
    return interpolated_image.reshape(*new_x.shape)



def data_noise_to_wavelet_light(lens_image, kwargs_res, model_type='source',
                                wavelet_type_list=['starlet', 'battle-lemarie-3'],
                                num_samples=1000, vmap_loop=True, sigma_clipping=True, seed=0,
                                starlet_num_scales=None, starlet_second_gen=False, 
                                noise_var=None, arc_mask=None, 
                                median_per_scale=False, delensing_type='operator'):
    # get the data noise
    nx, ny = lens_image.Grid.num_pixel_axes
    if noise_var is None:
        model_image = lens_image.model(**kwargs_res)
        diag_cov_d = lens_image.Noise.C_D_model(model_image)
    else:
        diag_cov_d = noise_var

    if np.any(np.where(diag_cov_d < 0.)):
        raise ValueError("Negative values in data covariance matrix")
    
    # number of source pixels
    if model_type == 'source':
        nx_out, ny_out = lens_image.SourceModel.pixel_grid.num_pixel_axes
    elif model_type == 'lens_light':
        nx_out, ny_out = lens_image.LensLightModel.pixel_grid.num_pixel_axes
    else:
        raise ValueError("This function only supports (pixelated) 'source' or 'lens_light' profiles")

    if model_type == 'source':
        if delensing_type == 'operator':
            # construct the lensing operator
            lensing_op = lens_image.get_lensing_operator(
                kwargs_lens=kwargs_res['kwargs_lens'], update=False, arc_mask=arc_mask,
            )
            def F_T(n): # de-lensing operation
                return lensing_op.lensing_transpose(n)
            
        elif delensing_type == 'interpol':
            if vmap_loop is True:
                raise NotImplementedError("Delensing operation via interpolation "
                                          "is not yet compatible with JAX's vmap")
            # TODO: update the following once it is possible to perform
            # interpolation on unstructured grids using JAX
            theta_x, theta_y = lens_image.Grid.pixel_coordinates 
            beta_prime_x, beta_prime_y = lens_image.SourceModel.pixel_grid.pixel_coordinates
            beta_x, beta_y = lens_image.MassModel.ray_shooting(theta_x, theta_y, kwargs_res['kwargs_lens'])
            def F_T(n):
                return interp_unstruct_grid(n, beta_x, beta_y, beta_prime_x, beta_prime_y)

    elif model_type == 'lens_light':
        def F_T(n): # identity operation
            return n
        
    # setup the transposed convolution
    kernel = jnp.copy(lens_image.PSF.kernel_point_source)
    nxk, nyk = kernel.shape
    kernel = kernel[:, :, jnp.newaxis, jnp.newaxis]
    dimension_numbers = ('NHWC', 'HWIO', 'NHWC')
    dn = lax.conv_dimension_numbers((1, nx, ny, 1),  # NHWC
                                    kernel.shape,  # HWIO
                                    dimension_numbers)
    kernel_rot = jnp.rot90(jnp.rot90(kernel, axes=(0, 1)), axes=(0, 1))
    
    def B_T(n): # transposed convolution
        res = lax.conv_general_dilated(n[jnp.newaxis, :, :, jnp.newaxis], 
                                       kernel_rot, 
                                       (1,1), #(k//2,k//2),  # window strides
                                       ((nxk//2, nyk//2), (nxk//2, nyk//2)), # padding mode
                                       (1,1), #(k//2,k//2),  # lhs/image dilation
                                       (1,1),  # rhs/kernel dilation
                                       dn)     # dimension_numbers = lhs, rhs, out dimension permutation
        return jnp.squeeze(res)

    wavelet_class_list = []
    std_per_scale_list = []

    for wavelet_type in wavelet_type_list:

        # setup the wavelet transform
        if 'battle-lemarie' in wavelet_type:
            nscales = 1  # we only care about the first scale for this one
        elif 'starlet' in wavelet_type:
            nscales_allowed = int(np.log2(min(nx_out, ny_out)))  # max number of scales allowed
            if starlet_num_scales is not None:
                nscales = min(nscales_allowed, starlet_num_scales)
            else:
                nscales = nscales_allowed

        wavelet = WaveletTransform(nscales, wavelet_type=wavelet_type, second_gen=starlet_second_gen)
        
        def Phi_T(n): # wavelet transform
            return wavelet.decompose(n)
        
        def propagate_noise(n):
            # apply linear operators to propagate noise from image plane to source plane
            tmp = F_T(B_T(n))
            if sigma_clipping is True:
                # here we clip values that are 5 times the standard deviation
                thresh = 5. * jnp.std(tmp)
                tmp = jnp.where(jnp.abs(tmp) > thresh, thresh, tmp)
            return Phi_T(tmp)

        # draw many realizations of the noise, scaled by the inverse cov matrix
        std_d_norm = 1. / np.sqrt(diag_cov_d) # = sigma * C_d^{-1}
        std_d_norm = std_d_norm[jnp.newaxis, :, :]
        noise_samples = std_d_norm * jax.random.normal(jax.random.PRNGKey(seed),
                                                       shape=(num_samples, nx, ny))

        # propagate the noise to wavelet space for each of them
        if vmap_loop is True:
            noise_samples_prop = vmap(jit(propagate_noise))(noise_samples)
        else:
            noise_samples_prop = []
            for noise in noise_samples:
                noise_samples_prop.append(propagate_noise(noise))
            noise_samples_prop = jnp.array(noise_samples_prop)

        # take the standard deviation
        std_per_scale = jnp.std(noise_samples_prop, axis=0)

        if median_per_scale is True:
            # single uniform value for each wavelet scale, we take the median value
            medians = jnp.nanmedian(std_per_scale, axis=(-2, -1))
            std_per_scale = np.full_like(std_per_scale, medians[:, np.newaxis, np.newaxis])
            
        wavelet_class_list.append(wavelet)
        std_per_scale_list.append(std_per_scale)

    return std_per_scale_list, wavelet_class_list



def data_noise_to_wavelet_potential(lens_image, kwargs_res, k_src=None,
                                    likelihood_type='chi2',
                                    wavelet_type_list=['starlet', 'battle-lemarie-3'], 
                                    starlet_second_gen=False,
                                    method='MC', num_samples=10000, seed=None, 
                                    ignore_poisson_noise=False, ignore_lens_light_flux=False,
                                    model_var_map=None, verbose=False):
    if likelihood_type not in ['l2_norm', 'chi2']:
        raise ValueError("Only 'l2_norm' and 'chi2' are supported options for likelihood_type.")

    mass_model = lens_image.MassModel
    kwargs_lens = kwargs_res['kwargs_lens']
    source_model = lens_image.SourceModel
    kwargs_source = kwargs_res['kwargs_source']

    # get model data variance
    model = lens_image.model(**kwargs_res, lens_light_add=(not ignore_lens_light_flux))
    if ignore_poisson_noise:
        data_var_map = lens_image.Noise.background_rms**2 * np.ones_like(model)
    else:
        data_var_map = lens_image.Noise.C_D_model(model, force_recompute=True)
    if model_var_map is not None:
        var_map = data_var_map + model_var_map  # add variances
    else:
        var_map = data_var_map

    var_map = np.array(var_map)  # cast to numpy array otherwise computations are slowed down
    std_map = np.sqrt(var_map)
    var_d = var_map.flatten()
    std_d = std_map.flatten()

    # extract coordinates grid, in image plane, ray-shot to source plane, and for the pixelated potential
    x_grid_d, y_grid_d = lens_image.Grid.pixel_coordinates
    x_grid_rs, y_grid_rs = mass_model.ray_shooting(x_grid_d, y_grid_d, kwargs_lens)
    x_grid_psi, y_grid_psi = lens_image.MassModel.pixel_grid.pixel_coordinates

    # number of pixels and wavelet scales
    nx_d, ny_d = x_grid_d.shape
    nx_psi, ny_psi = x_grid_psi.shape
    
    # compute derivatives of the source light at ray-shot coordinates
    source_deriv_x, source_deriv_y = source_model.spatial_derivatives(x_grid_rs, 
                                                                      y_grid_rs, 
                                                                      kwargs_source, k=k_src)
    source_deriv_x *= lens_image.Grid.pixel_area  # correct flux units
    source_deriv_y *= lens_image.Grid.pixel_area  # correct flux units

    # reshape all quantities for building the operator
    data_x = x_grid_d.flatten()
    data_y = np.flip(y_grid_d, axis=0).flatten()  # WARNING: y-coords must be flipped vertically!
    dpsi_x = x_grid_psi.flatten()
    dpsi_y = np.flip(y_grid_psi, axis=0).flatten()  # WARNING: y-coords must be flipped vertically!
    dpsi_dx = abs(x_grid_psi[0, 0] - x_grid_psi[0, 1])
    dpsi_dy = abs(y_grid_psi[0, 0] - y_grid_psi[1, 0])
    dpsi_xmin = x_grid_psi[0, :].min() - dpsi_dx/2.
    dpsi_ymax = y_grid_psi[:, 0].max() + dpsi_dy/2.
    dpsi_Nx, dpsi_Ny = nx_psi, ny_psi
    source0_dx = source_deriv_x.flatten()
    source0_dy = source_deriv_y.flatten()

    # get the DsD operator (see Koopmans+05)
    start = time.time()
    DsDpsi_matrix = vkl_util.vkl_operator(np.array(data_x), np.array(data_y),
                                          dpsi_xmin, dpsi_dx, dpsi_Nx, np.array(dpsi_x), 
                                          dpsi_ymax, dpsi_dy, dpsi_Ny, np.array(dpsi_y),
                                          np.array(source0_dx), np.array(source0_dy))
    if verbose: print("compute DsDpsi:", time.time()-start)

    # blurring operator for PSF convolutions
    start = time.time()
    B_matrix = lens_image.PSF.blurring_matrix(model.shape)
    if verbose: print("compute B:", time.time()-start)

    # D operator
    start = time.time()
    D_matrix  = - B_matrix.dot(DsDpsi_matrix)
    DT_matrix = D_matrix.T
    if verbose: print("compute D:", time.time()-start)

    # wavelet transform Phi^T operators
    PhiT_operator_list = []
    num_scales_list = []
    for wavelet_type in wavelet_type_list:
        if wavelet_type in ['battle-lemarie-1', 'battle-lemarie-3']:
            num_scales = 1  # we only care about the first scale for this one
        else:
            num_scales = int(np.log2(min(nx_psi, ny_psi)))
        wavelet = WaveletTransform(num_scales, wavelet_type=wavelet_type, 
                                   second_gen=starlet_second_gen)
        PhiT_operator_list.append(wavelet.decompose)
        num_scales_list.append(num_scales)


    if method == 'MC':

        # initialize random generator seed
        np.random.seed(seed)

        # draw samples from the data covariance matrix
        # cov_d = np.diag(var_d)
        # mu = np.zeros(nx_d*ny_d)  # zero mean
        # noise_reals = np.random.multivariate_normal(mu, cov_d, size=num_samples)
        
        # apply the operators for each realization of the noise
        psi_wt_std_list = []
        for wavelet_type, PhiT_operator in zip(wavelet_type_list, PhiT_operator_list):

            start = time.time()
            psi_wt_reals = []
            for i in range(num_samples):
                # noise_i = noise_reals[i, :]
                noise_i = std_d * np.random.randn(*std_d.shape)  # draw a noise realization

                # if chi2 loss, rescale by the data variance
                if likelihood_type == 'chi2':
                    noise_i /= var_d

                psi_i = DT_matrix.dot(noise_i)
                psi_wt_i = PhiT_operator( psi_i.reshape(nx_psi, ny_psi) )
                psi_wt_reals.append(psi_wt_i)
            psi_wt_reals = np.array(psi_wt_reals) # --> shape = (num_samples, num_scales, nx_psi, ny_psi)
            if verbose: print(f"loop over MC samples for wavelet '{wavelet_type}':", time.time()-start)

            # compute the variance per wavelet scale per potential pixel over all the samples
            psi_wt_var = np.var(psi_wt_reals, axis=0)
            # check
            if np.any(psi_wt_var < 0.):
                raise ValueError("Negative variance terms!")

            # convert to standard deviation
            psi_wt_std = np.sqrt(psi_wt_var)

            psi_wt_std_list.append(psi_wt_std)


    elif method == 'SLIT':

        psi_wt_std_list = []
        for PhiT_operator, num_scales in zip(PhiT_operator_list, num_scales_list):

            # following the same recipe as SLIT(ronomy):

            # first term of
            B_noise = std_map * np.sqrt(np.sum(psf_kernel_2d.T**2))
            B_noise = B_noise.flatten()
            DsDpsiB_noise = (DsDpsi_matrix.T).dot(B_noise)
            DsDpsiB_noise = DsDpsiB_noise.reshape(nx_psi, ny_psi)

            dirac = np.zeros((nx_psi, ny_psi))
            if nx_psi % 2 == 0:
                warnings.warn("Expect the potential noise maps to be shifted by one pixel (psi grid has even size!).")
            dirac[nx_psi//2, ny_psi//2] = 1
            dirac_wt = PhiT_operator(dirac)

            psi_wt_std = []
            for k in range(num_scales+1):
                psi_wt_std2_k = signal.fftconvolve(DsDpsiB_noise**2, dirac_wt[k]**2, mode='same')
                psi_wt_std_k = np.sqrt(psi_wt_std2_k)
                psi_wt_std.append(psi_wt_std_k)
            psi_wt_std = np.array(psi_wt_std) # --> shape = (num_scales, nx_psi, ny_psi)

            psi_wt_std_list.append(psi_wt_std)

    else:
        raise ValueError(f"Method '{method}' for noise propagation is not supported.")
    
    return psi_wt_std_list



# Describes a collection of mass models used for multi-plane lensing
# 
# Copyright (c) 2024, herculens developers and contributors

__author__ = 'krawczyk'

import numpy as np
import jax
import jax.numpy as jnp

from functools import partial
from herculens.MassModel.mass_model import MassModel, ZeroMassModel



# def ray_shooting_static_for_scan(
#         alpha_list,
#         eta_matrix_t,
#         carry,
#         j,
#     ):
#     xs, ys = carry
#     dx, dy = jax.lax.switch(j, alpha_list, xs[j], ys[j])
#     etas_j = jax.lax.dynamic_slice(eta_matrix_t, (0, j), (len(alpha_list)+1, 1))
#     xs = xs - etas_j * dx
#     ys = ys - etas_j * dy
#     return (xs, ys), None


class MPMassModel(object):
    def __init__(
            self, 
            mp_mass_model_list, 
            deflection_scaling_convention='standard', 
            **mass_model_kwargs,
        ):
        '''
        Create a MPMassModel object.

        Parameters
        ----------
        mp_mass_model_list : list of str
            List of lists containing Lens model profiles for each plane of
            the lens system. One inner list per plane with the outer list
            sorted by distance from observer.
            If a mass plane has no mass model associated to it, use None.
        mass_model_kwargs : dictionary for settings related to PIXELATED
            profiles.
        deflection_scaling_convention : str, optional
            Either 'standard' or 'glee'. Determines the convention used for
            the eta matrix. The default is 'standard' (see build_eta_matrix() docstring). 
            If 'glee' is used, the eta matrix is built by assuming additional
            distance ratio factors as is done in the GLEE lens modelling software. 
            The default is 'standard'.
        '''
        string_input = all([isinstance(mm, str) or mm is None for mm in mp_mass_model_list])
        instance_input = all([isinstance(mm, MassModel) or mm is None for mm in mp_mass_model_list])
        if deflection_scaling_convention.lower() not in ('standard', 'glee'):
            raise ValueError(
                "MPMassModel convention must be either 'standard' or 'glee'."
            )
        else:
            self._scaling_conv = deflection_scaling_convention.lower()
        zero_mass = ZeroMassModel()
        if string_input:
            self.mass_models = [MassModel(mm, **mass_model_kwargs) if mm is not None else zero_mass for mm in mp_mass_model_list]
            self.mp_profile_type_list = mp_mass_model_list
        elif instance_input:
            self.mass_models = [mm if mm is not None else zero_mass for mm in mp_mass_model_list]
            self.mp_profile_type_list = [mm.func_list if mm is not None else None for mm in self.mass_models]
        else:
            raise ValueError(
                "MPMassModel needs to be initialized either with a list of lists of strings (or None), "
                "or directly with a list of (single plane) MassModel instances (or None)."
            )
        self.number_mass_planes = len(self.mass_models)
        if self._scaling_conv == 'glee':
            # Eta will be passed in flattened to `ray_shooting`, use these
            # index values to un-flatten it back into an array
            self.eta_idx = np.triu_indices(self.number_mass_planes + 1, k=1)
            # The know values for the un-flattened eta array
            self.base_eta = jnp.zeros((self.number_mass_planes + 1, self.number_mass_planes + 1))
        else:
            # Eta will be passed in flattened to `ray_shooting`, use these
            # index values to un-flatten it back into an array
            self.eta_idx = np.triu_indices(self.number_mass_planes + 1, k=2)
            # The know values for the un-flattened eta array
            self.base_eta = jnp.eye(self.number_mass_planes + 1, k=1)

    @property
    def has_pixels(self):
        # An list of bools indicating if a plane contains an pixelated mass model
        if not hasattr(self, '_has_pixels'):
            self._has_pixels = [mass_model.has_pixels if mass_model is not None else False for mass_model in self.mass_models]
        return self._has_pixels
    
    @property
    def has_mass(self):
        if not hasattr(self, '_has_mass'):
            # An list of bools indicating if a plane contains a mass model
            self._has_mass = [not isinstance(mass_model, ZeroMassModel) for mass_model in self.mass_models]
        return self._has_mass
    
    @property
    def indices_planes_with_mass(self):
        if not hasattr(self, '_indices_planes_with_mass'):
            # An list of bools indicating if a plane contains a mass model
            self._indices_planes_with_mass = [i for i in range(self.number_mass_planes) if self.has_mass[i]]
        return self._indices_planes_with_mass

    @partial(jax.jit, static_argnums=(0, 5, 6))
    def ray_shooting(self, x, y, eta_flat, kwargs, N=None, k=None):
        '''Maps image to source position (inverse deflection) on each mass plane

        Parameters
        ----------
        x : jax.numpy array
            x-position (preferentially arcsec)
        y : jax.numpy array
            y-position (preferentially arcsec)
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in the stack).
        kwargs: list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        N : int, optional
            number of planes to ray trace (front to back, starting at 1), by default None
        k : list of list, optional
            only evaluate the k-th lens model (list of list of index values) for a particular
            plane, by default None

        Returns
        -------
        x_deflected : jax.numpy array
            x source plane positions on each mass plane (first index of each corresponds to
            mass plane)
        y_deflected : jax.numpy array
            y source plane positions on each mass plane (first index of each corresponds to
            mass plane)
        '''
        if (N is None) or (N > self.number_mass_planes):
            N = self.number_mass_planes
        if k is None:
            k = [None] * N
        # un-flatten eta_flat into full eta array
        etas_t = self.base_eta.at[self.eta_idx].set(eta_flat)[:-1, :(N + 1)].T
        
        # create output array
        xs = jnp.stack([x] * (N + 1), axis=0)
        ys = jnp.stack([y] * (N + 1), axis=0)

        # iterate the lensing equation for each mass plane
        for j in range(N):
            dx, dy = self.mass_models[j].alpha(
                xs[j],
                ys[j],
                kwargs=kwargs[j],
                k=k[j]
            )
            etas_j = etas_t[:, j:j + 1]
            xs = xs - etas_j * dx
            ys = ys - etas_j * dy

        # WIP: the following code is better jax code but can still be improved
        # alpha_list = [
        #     partial(
        #         self.mass_models[j].alpha,
        #         kwargs=kwargs[j],
        #         k=k[j],
        #     ) for j in range(N)
        # ]
        # # recursive function with keywords filled in
        # partial_ray_shooting = partial(
        #     ray_shooting_static_for_scan,
        #     alpha_list,
        #     etas_t,
        # )
        # # run recursion function
        # xs = jnp.stack([x] * (N + 1), axis=0)
        # ys = jnp.stack([y] * (N + 1), axis=0)
        # (xs, ys), _ = jax.lax.scan(
        #     partial_ray_shooting,
        #     (xs, ys),
        #     jnp.arange(N, dtype=int),
        # )

        return xs, ys

    def _ray_shooting_slice(self, x, y, eta_flat, kwargs):
        '''Helper function that give *scalar* inputs of x and y give a *vector*
        output for each mass plane. Used for the vectorization of the `A` method'''
        return jnp.stack(
            self.ray_shooting(jnp.array([x]), jnp.array([y]), eta_flat, kwargs)
        ).T.squeeze()

    def _A_stack(self, x, y, eta_flat, kwargs, kind='auto'):
        '''Helper function that takes the jacobian of the ray shooting give *scalar*
        inputs for x and y and returns a 2x2 array.'''
        if kind == 'auto':
            return jnp.stack(
                jax.jacfwd(
                    self._ray_shooting_slice,
                    argnums=(0, 1)
                )(x, y, eta_flat, kwargs)
            )
        elif kind == 'direct':
            N = self.number_mass_planes
            A = jnp.kron(jnp.eye(2), jnp.ones((N + 1, 1))).reshape(2, N + 1, 2)
            xs, ys = self.ray_shooting(jnp.array([x]), jnp.array([y]), eta_flat, kwargs)
            etas_t = self.base_eta.at[self.eta_idx].set(eta_flat)[:-1, :(N + 1)].T
            for j in range(N):
                if self.has_mass[j]:
                    hxx, hxy, hyx, hyy = self.mass_models[j].hessian(
                        xs[j][0],
                        ys[j][0],
                        kwargs[j]
                    )
                else:
                    hxx, hxy, hyx, hyy = 0., 0., 0., 0.
                etas_j = etas_t[:, j:j + 1]
                A_H = A[:, j, :] @ jnp.array([[hxx, hxy], [hyx, hyy]])
                A = A - jnp.kron(A_H, etas_j).reshape(2, N + 1, 2)
            return A

    @partial(jax.jit, static_argnums=(0, 5))
    def A(self, x, y, eta_flat, kwargs, kind='auto'):
        '''
        Area distortion matrix of the lens mapping.

        Parameters
        ----------
        x : jax.numpy array
            x-position (preferentially arcsec)
        y : jax.numpy array
            y-position (preferentially arcsec)
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
        kwargs: list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        kind : str, optional
            either "auto" or "direct". Determines how the distortion matrix is
            computed, "auto" will using automatic differentiation using using JAX's
            `jaxfwd` function. "direct" will using the `hessian` method for each
            mass plane. "auto" is typically faster and is the default and recommended
            method.

        Returns
        -------
        jax.numpy array
            The area distortion matrix of the lens for each position
            and each mass plane (including the image plane) with shape
            (N+1, *(x.shape), 2, 2) where N is the number of mass planes.
        '''
        A_stack_part = partial(
            self._A_stack,
            eta_flat=eta_flat,
            kwargs=kwargs,
            kind=kind
        )
        return jnp.moveaxis(jnp.vectorize(
            A_stack_part,
            signature='(),()->(i,j,i)'
        )(x, y), -2, 0)

    def inverse_magnification(self, x, y, eta_flat, kwargs, kind='auto'):
        '''Return the inverse magnification map for each plane of the lens
        system.

        Parameters
        ----------
        x : jax.numpy array
            x-position (preferentially arcsec)
        y : jax.numpy array
            x-position (preferentially arcsec)
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
        kwargs : list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        kind : str, optional
            either "auto" or "direct". Determines how the distortion matrix is
            computed, "auto" will using automatic differentiation using using JAX's
            `jaxfwd` function. "direct" will using the `hessian` method for each
            mass plane. "auto" is typically faster and is the default and recommended
            method.

        Returns
        -------
        jax.numpy array
            The inverse magnification of the lens for each position
            and each mass plane (including the image plane) with shape
            (N+1, *(x.shape)) where N is the number of mass planes.
        '''
        A = self.A(x, y, eta_flat, kwargs, kind=kind)
        return A[..., 0, 0] * A[..., 1, 1] - A[..., 0, 1] * A[..., 1, 0]
    
    def magnification(self, x, y, eta_flat, kwargs, kind='auto'):
        '''Return the magnification map for each plane of the lens
        system.

        Parameters
        ----------
        x : jax.numpy array
            x-position (preferably arcsec)
        y : jax.numpy array
            x-position (preferably arcsec)
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
        kwargs : list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        kind : str, optional
            either "auto" or "direct". Determines how the distortion matrix is
            computed, "auto" will using automatic differentiation using using JAX's
            `jaxfwd` function. "direct" will using the `hessian` method for each
            mass plane. "auto" is typically faster and is the default and recommended
            method.

        Returns
        -------
        jax.numpy array
            The magnification of the lens for each position and each mass plane (including the
            image plane) with shape (N+1, *(x.shape)) where N is the number of mass planes.
        '''
        return 1. / self.inverse_magnification(x, y, eta_flat, kwargs, kind=kind)

    def kappa(self, x, y, eta_flat, kwargs, kind='auto'):
        '''Lensing convergence k = 1/2 laplacian(phi) map for each plane of the lens
        system.

        Parameters
        ----------
        x : jax.numpy array
            x-position (preferentially arcsec)
        y : jax.numpy array
            x-position (preferentially arcsec)
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
        kwargs : list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        kind : str, optional
            either "auto" or "direct". Determines how the distortion matrix is
            computed, "auto" will using automatic differentiation using using JAX's
            `jaxfwd` function. "direct" will using the `hessian` method for each
            mass plane. "auto" is typically faster and is the default and recommended
            method.

        Returns
        -------
        jax.numpy array
            The lensing convergence for each position and each mass plane (including the
            image plane) with shape (N+1, *(x.shape)) where N is the number of mass planes.
        '''
        A = self.A(x, y, eta_flat, kwargs, kind=kind)
        return 1 - 0.5 * (A[..., 0, 0] + A[..., 1, 1])

    def gamma(self, x, y, eta_flat, kwargs, kind='auto'):
        '''shear computation for each plane of the lens system
        g1 = 1/2(d^2phi/dx^2 - d^2phi/dy^2)
        g2 = d^2phi/dxdy

        Parameters
        ----------
        x : jax.numpy array
            x-position (preferentially arcsec)
        y : jax.numpy array
            x-position (preferentially arcsec)
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
        kwargs : list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        kind : str, optional
            either "auto" or "direct". Determines how the distortion matrix is
            computed, "auto" will using automatic differentiation using using JAX's
            `jaxfwd` function. "direct" will using the `hessian` method for each
            mass plane. "auto" is typically faster and is the default and recommended
            method.

        Returns
        -------
        gamma1 : jax.numpy array
            The first shear component for each position and each mass plane (including the
            image plane) with shape (N+1, *(x.shape)) where N is the number of mass planes.
        gamma2 : jax.numpy array
            The second shear component for each position and each mass plane (including the
            image plane) with shape (N+1, *(x.shape)) where N is the number of mass planes.
        '''
        A = self.A(x, y, eta_flat, kwargs, kind=kind)
        gamma1 = 0.5 * (A[..., 1, 1] - A[..., 0, 0])
        gamma2 = -A[..., 0, 1]
        return gamma1, gamma2
    
    def build_eta_matrix(
            self,
            cosmology, 
            redshifts, 
            return_matrix=False, 
            return_labels=False,
        ):
        """Utility function to build the eta matrix with the right conventions
        for use in a multi-plane lens model.

        For more details, see comments in the Pull Request #37:
        https://github.com/Herculens/herculens/pull/37#issuecomment-2343179184

        Note however that when deflection_scaling_convention is set to 'glee', the
        eta matrix is built by assuming the GLEE convention, which contains
        additional scaling factors, including in the upper diagonal elements of the matrix.

        Parameters
        ----------
        cosmology : Astropy cosmology
            Instance of an Astropy cosmology (e.g. `LambdaCDM`).
        redshifts : list or ndarray
            List or 1D array containing the redshift for each plane, starting from the lowest redshift one.
        return_matrix : bool
            If True, returns also the full matrix which includes
            trivial elements of the matrix (0s and 1s). Default is False.
        return_labels : bool
            If True, returns the strings for each eta element that detail
            the distance ratios and their individual values. Default is False.
        """
        # check that the redshifts are sorted
        np.testing.assert_equal(np.sort(redshifts), redshifts)
        # check that the redshifts are unique
        np.testing.assert_equal(np.unique(redshifts), redshifts)
        if self._scaling_conv == 'glee':
            return self._build_eta_matrix_glee(
                cosmology, 
                redshifts, 
                return_matrix=return_matrix, 
                return_labels=return_labels,
            )
        else:
            return self._build_eta_matrix_std(
                cosmology, 
                redshifts, 
                return_matrix=return_matrix, 
                return_labels=return_labels,
            )
    
    def eta_flat_to_matrix(self, eta_flat):
        """Un-flatten the eta matrix from the flat version to the full version
        (including the trivial elements of the matrix)."""
        #N = self.number_mass_planes
        # un-flatten eta_flat into full eta array
        return self.base_eta.at[self.eta_idx].set(eta_flat) #[:-1, :(N + 1)]
    
    def _build_eta_matrix_std(
            self,
            cosmology, 
            redshifts, 
            return_matrix=False, 
            return_labels=False,
        ):
        # iterate over the planes
        num_tot_planes = len(redshifts)
        eta_flat = []
        eta_labels = []
        for i in range(num_tot_planes):
            for j in range(i+2, num_tot_planes):

                z_i = redshifts[i]
                z_ip1 = redshifts[i+1]
                z_j  = redshifts[j]

                D_j  = cosmology.angular_diameter_distance(z_j).value
                D_i_j = cosmology.angular_diameter_distance_z1z2(z_i, z_j).value
                D_ip1 = cosmology.angular_diameter_distance(z_ip1).value
                D_i_ip1 = cosmology.angular_diameter_distance_z1z2(z_i, z_ip1).value

                eta_ij = (D_i_j * D_ip1) / (D_j * D_i_ip1)

                eta_labels.append(
                    f"eta_{i}_{j} "
                    f"= ( D_{i}_{j}({z_i:.3f}, {z_j:.3f}) x D_{i+1}(0, {z_ip1:.3f}) ) / ( D_{j}(0, {z_j:.3f}) x D_{i}_{i+1}({z_i:.3f}, {z_ip1:.3f}) ) "
                    f"= ( {D_i_j:.1f} x {D_ip1:.1f} ) / ( {D_j:.1f} x {D_i_ip1:.1f} ) "
                    f"= {eta_ij:.2f}"
                )
                # print(f"eta_{i:03}_{j:03}", eta_ij)

                eta_flat.append(eta_ij)

        eta_flat = np.array(eta_flat)

        if return_matrix is False:
            if return_labels is True:
                return eta_flat, eta_labels
            else:
                return eta_flat
        
        eta_full = self.eta_flat_to_matrix(eta_flat)

        if return_labels is True:
            return eta_flat, eta_full, eta_labels
        else:
            return eta_flat, eta_full
        
    def _build_eta_matrix_glee(
            self,
            cosmology, 
            redshifts, 
            return_matrix=False, 
            return_labels=False,
        ):
        """
        TODO
        """
        # iterate over the planes
        num_tot_planes = len(redshifts)
        eta_flat = []
        eta_labels = []
        for i in range(num_tot_planes-1):
            z_i = redshifts[i]
            z_ip1 = redshifts[i+1]

            # D_i = cosmology.angular_diameter_distance(z_i).value
            D_ip1 = cosmology.angular_diameter_distance(z_ip1).value
            D_i_ip1 = cosmology.angular_diameter_distance_z1z2(z_i, z_ip1).value

            eta_i_ip1 = (D_i_ip1 / D_ip1)
            eta_flat.append(eta_i_ip1)
            
            for j in range(i+2, num_tot_planes):
                z_j  = redshifts[j]

                # print(f"z_i: {z_i}, z_ip1: {z_ip1}, z_j: {z_j}")

                D_j  = cosmology.angular_diameter_distance(z_j).value
                D_i_j = cosmology.angular_diameter_distance_z1z2(z_i, z_j).value

                # print("(D_i_j / D_j)", (D_i_j / D_j))

                # eta_ij = (D_i_j * D_ip1) / (D_j * D_i_ip1) * (D_i_j / D_j)
                
                eta_ij = (D_i_j / D_j)
                
                eta_labels.append(
                    "N/A"  # TODO
                )
                # print(f"eta_{i:03}_{j:03}", eta_ij)

                eta_flat.append(eta_ij)

        eta_flat = np.array(eta_flat)

        if return_matrix is False:
            if return_labels is True:
                return eta_flat, eta_labels
            else:
                return eta_flat
        
        eta_full = self.eta_flat_to_matrix(eta_flat)

        if return_labels is True:
            return eta_flat, eta_full, eta_labels
        else:
            return eta_flat, eta_full
        
    


# Handles different method to optimize a loss function
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal', 'austinpeel'


import time
from copy import deepcopy
import jax
import jaxopt

from herculens.Inference.Optimization.base_optim import BaseOptimizer


__all__ = ['JaxoptOptimizer']


class JaxoptOptimizer(BaseOptimizer):
    """Wrapper to jaxopt's unconstrained optimizers"""

    def run_scipy(self, init_params, progress_bar=True, 
                  return_param_history=False, **solver_kwargs):
        # TODO: should we call once / a few times all jitted functions before optimization, to potentially speed things up?
        metrics = MinimizeMetrics(self.loss.function, with_param_history=return_param_history)
        solver = jaxopt.ScipyMinimize(fun=self.function_optim, jit=True, 
                                      callback=metrics, **solver_kwargs)

        init_params_ = deepcopy(init_params)
        start = time.time()
        # runs the optimizer
        res = solver.run(init_params_)
        # retrieve optimized parameters and loss value
        best_fit = res.params
        logL_best_fit = - self.loss.function(best_fit)
        runtime = time.time() - start

        extra_fields = {}
        extra_fields['loss_history'] = metrics.get_loss_history()
        if return_param_history is True:
            extra_fields['param_history'] = metrics.get_param_history()
        return best_fit, logL_best_fit, extra_fields, runtime

    def run(self, init_params, method='BFGS', progress_bar=False, **solver_kwargs):
        if method == 'BFGS':
            solver = jaxopt.BFGS(self.function_optim, value_and_grad=False, 
                                 **solver_kwargs)
        elif method == 'LBFGS':
            solver = jaxopt.LBFGS(self.function_optim, value_and_grad=False, 
                                  **solver_kwargs)
        elif method == 'LM':
            solver = jaxopt.LevenbergMarquardt(self.function_optim_LM_scalar, 
                                               **solver_kwargs)
        else:
            raise NotImplementedError

        # Defines and jits a single solver update
        @jax.jit
        def step(params_state, _):
            params, state = params_state
            params, state = solver.update(params, state)
            loss_val = self.loss.function(params)
            return (params, state), loss_val

        # Initialise optimizer state
        params = deepcopy(init_params)
        state = solver.init_state(params)

        # Gradient descent loop
        maxiter = solver_kwargs.pop('maxiter')

        start_time = time.time()
        if progress_bar:
            # param_history = []
            loss_history = []
            for i in self._for_loop(range(maxiter), progress_bar, 
                                    total=maxiter, 
                                    desc=f"jaxopt.{method}"):
                (params, state), loss_val = step((params, state), None)
                loss_history.append(loss_val)
                # if return_param_history is True:
                #     param_history.append(params)
        else:
            (params, state), loss_history = jax.lax.scan(step, (params, state), None, length=maxiter)
        runtime = time.time() - start_time

        # start_time = time.time()
        # init_params_ = deepcopy(init_params)
        # params, state = solver.run(init_params_)
        # runtime = time.time() - start_time

        best_fit = params
        logL_best_fit = - self.loss.function(best_fit)
        extra_fields = {'loss_history': loss_history}
        # if return_param_history is True:
        #     extra_fields['param_history'] = param_history
        return best_fit, logL_best_fit, extra_fields, runtime


class MinimizeMetrics(object):
    """simple callable class used as callback in scipy.optimize.minimize method"""
    
    def __init__(self, func, with_param_history=False):
        self._func = func
        self._call = self._call_1arg
        self._with_param_history = with_param_history
        self.reset()

    def reset(self):
        self.loss_history = []
        self.param_history = []

    def get_loss_history(self, copy=True):
        return deepcopy(self.loss_history) if copy else self.loss_history

    def get_param_history(self, copy=True):
        return deepcopy(self.param_history) if copy else self.param_history

    def __call__(self, *args, **kwargs):
        return self._call(*args, **kwargs)
        
    def _call_1arg(self, x):
        self.loss_history.append(float(self._func(x)))
        if self._with_param_history:
            self.param_history.append(x)

    def _call_2args(self, x, state):
        # Input state parameter is necessary for 'trust-constr' method
        # You can use it to stop execution early by returning True
        self.loss_history.append(float(self._func(x)))
        if self._with_param_history:
            self.param_history.append(x)



# Describes a collection of light models used for multi-plane lensing
# 
# Copyright (c) 2024, herculens developers and contributors

__author__ = 'krawczyk'

import jax
import jax.numpy as jnp

from functools import partial
from herculens.LightModel.light_model import LightModel


class MPLightModel(object):
    @staticmethod
    def partial_flux(light_model, x, y, kwargs, k):
        return light_model.surface_brightness(x, y, kwargs, k)

    @staticmethod
    def partial_spatial_derivative(light_model, x, y, kwargs, k):
        return light_model.spatial_derivatives(x, y, kwargs, k)

    def __init__(self, mp_light_model_list, **light_model_kwargs):
        '''
        Create a MPLightModel object.

        Parameters
        ----------
        mp_light_model_list : list of str
            List of lists containing light model profiles for each plane of
            the lens system. One inner list per plane with the outer list
            sorted by distance from observer starting with the lens light.
        light_model_kwargs : dictionary for settings related to PIXELATED
            profiles.
        '''
        string_input = all([isinstance(lm, str) or lm is None for lm in mp_light_model_list])
        instance_input = all([isinstance(lm, LightModel) or lm is None for lm in mp_light_model_list])
        if string_input:
            self.light_models = [LightModel(lm, **light_model_kwargs) if lm is None else None for lm in mp_light_model_list]
            self.mp_profile_type_list = mp_light_model_list
        elif instance_input:
            self.light_models = mp_light_model_list
            self.mp_profile_type_list = [lm.func_list if lm is not None else None for lm in self.light_models]
        else:
            raise ValueError(
                "MPLightModel needs to be initialized either with a list of lists of strings, "
                "or directly with a list of (single plane) LightModel instances."
            )
        self.number_light_planes = len(self.mp_profile_type_list)

    @property
    def has_pixels(self):
        return [light_model.has_pixels if light_model is not None else False for light_model in self.light_models]
    
    @property
    def has_light(self):
        return [light_model is not None for light_model in self.light_models]

    @property
    def pixel_is_adaptive(self):
        return [light_model.pixel_is_adaptive if light_model is not None else False for light_model in self.light_models]

    @property
    def pixelated_shape(self):
        return [light_model.pixelated_shape if light_model is not None else None for light_model in self.light_models]

    @property
    def pixelated_coordinates(self):
        return [light_model.pixelated_coordinates if light_model is not None else None for light_model in self.light_models]

    @property
    def pixelated_index(self):
        return [light_model.pixelated_index if light_model is not None else None for light_model in self.light_models]

    @property
    def pixel_grid(self):
        return [light_model.pixel_grid if light_model is not None else None for light_model in self.light_models]

    @property
    def pixel_grid_settings(self):
        return [light_model.pixel_grid_settings if light_model is not None else None for light_model in self.light_models]

    @property
    def param_name_list(self):
        return [light_model.param_name_list if light_model is not None else None for light_model in self.light_models]

    def k_expand(self, k):
        # helper function to expand `None` into a list of `None` values of the
        # expected length.
        if k is None:
            return [None] * self.number_light_planes
        else:
            return k

    @partial(jax.jit, static_argnums=(0, 6))
    def surface_brightness(
        self,
        x,
        y,
        kwargs_list,
        pixels_x_coord,
        pixels_y_coord,
        k=None,
    ):
        '''Total source flux at a given position.

        Parameters
        ----------
        x, y : array_like
            Position coordinate(s) in arcsec relative to the image center for each light
            plane (first index corresponds to light plane).
        kwargs_list : list
            List of lists of parameter dictionaries corresponding to each source model.
        pixel_x_coords, pixel_y_coords : array_like
            Pixel position coordinate(s) for an adaptive grid based on arc masks (see MPLensImage).
        k : list of list, optional
            only evaluate the k-th light model (list of list of index values) for each light
            plane, by default None

        Returns
        -------
        array_like
            Flux for each image plane in the lens system (first index of each corresponds to
            light plane).
        '''
        k = self.k_expand(k)
        flux = []
        for j in range(self.number_light_planes):
            if self.has_light[j]:
                flux.append(
                    self.light_models[j].surface_brightness(
                        x[j], y[j], kwargs_list[j],
                        k=k[j],
                        pixels_x_coord=pixels_x_coord[j],
                        pixels_y_coord=pixels_y_coord[j]
                    )
                )
        return jnp.stack(flux)

    def spatial_derivatives(self, x, y, kwargs_list, k=None):
        '''Spatial derivatives of the source flux at a given position (along x and y directions)
        for each light plane.

        Parameters
        ----------
        x, y : array_like
            Position coordinate(s) in arcsec relative to the image center for each light
            plane (first index corresponds to light plane).
        kwargs_list : list
            List of lists of parameter dictionaries corresponding to each source model.
        list of list, optional
            only evaluate the k-th light model (list of list of index values) for each light
            plane, by default None

        Returns
        -------
        array_like
            Spatial derivatives of the source flux at a given position for each light plane
            (first index of each corresponds to light plane).
        '''
        k = self.k_expand(k)
        results = []
        for j in range(self.number_light_planes):
            results.append(self.light_models[j].spatial_derivatives(
                x[j], y[j], kwargs_list[j], k[j]
            ))
        return jnp.stack(results)




# High-level interface to a mass model
#
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


from functools import partial
import numpy as np
import jax.numpy as jnp
import jax
from jax import jit, lax
from jax.tree_util import register_pytree_node

from herculens.MassModel.mass_model_base import MassModelBase
from herculens.MassModel.Profiles.zero_mass import ZeroMass


__all__ = ['MassModel', 'ZeroMassModel']


def tree_stack(trees):
    return jax.tree.map(lambda *v: jnp.stack(v), *trees)


def alpha_static_single(
        x, y, alpha_func,
    ):
    return partial(alpha_func, x, y)

# @partial(jit, static_argnums=(0, 4))
def alpha_static_for_scan(
        x, y,
        alpha_list,
        carry,
        j
    ):
    f_x, f_y = carry
    f_x_j, f_y_j = jax.lax.switch(j, alpha_list, x, y)
    f_x = f_x + f_x_j
    f_y = f_y + f_y_j
    return (f_x, f_y), None


class MassModel(MassModelBase):
    """An arbitrary list of lens models."""
    def __init__(self, profile_list, use_jax_scan=False, verbose=False, **kwargs):
        """Create a MassModel object.

        Parameters
        ----------
        profile_list : list of strings or profile instances
            List of mass profiles.
        use_jax_scan : bool
            If True, uses jax.lax.scan to evaluate deflection angles, which may speed up compilation and run time for a large number of mass profiles.
        kwargs : dict
            See docstring for MassModelBase.get_class_from_string()
        """
        if not isinstance(profile_list, (list, tuple)):
            # useful when using a single profile
            self.profile_type_list = [profile_list]
        else:
            self.profile_type_list = profile_list
        super().__init__(self.profile_type_list, **kwargs)
        self._use_jax_scan = use_jax_scan
        self._repeated_profile_mode = False
        if len(self.profile_type_list) > 0:
            first_profile = self.profile_type_list[0]
            self._repeated_profile_mode = (
                all(p is first_profile for p in self.profile_type_list)
            )
            if verbose is True and self._repeated_profile_mode:
                print("All MassModel profiles are identical.")

    @partial(jit, static_argnums=(0, 4))
    def ray_shooting(self, x, y, kwargs, k=None):
        """
        maps image to source position (inverse deflection)
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: source plane positions corresponding to (x, y) in the image plane
        """
        dx, dy = self.alpha(x, y, kwargs, k=k)
        return x - dx, y - dy

    def fermat_potential(self, x_image, y_image, kwargs_lens, x_source=None, y_source=None, k=None):
        """
        fermat potential (negative sign means earlier arrival time)

        :param x_image: image position
        :param y_image: image position
        :param x_source: source position
        :param y_source: source position
        :param kwargs_lens: list of keyword arguments of lens model parameters matching the lens model classes
        :return: fermat potential in arcsec**2 without geometry term (second part of Eqn 1 in Suyu et al. 2013) as a list
        """
        potential = self.potential(x_image, y_image, kwargs_lens, k=k)
        if x_source is None or y_source is None:
            x_source, y_source = self.ray_shooting(x_image, y_image, kwargs_lens, k=k)
        geometry = ((x_image - x_source)**2 + (y_image - y_source)**2) / 2.
        return geometry - potential

    def potential(self, x, y, kwargs, k=None):
        """
        lensing potential
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing potential in units of arcsec^2
        """
        # x = np.array(x, dtype=float)
        # y = np.array(y, dtype=float)
        if isinstance(k, int):
            return self.func_list[k].function(x, y, **kwargs[k])
        bool_list = self._bool_list(k)
        potential = np.zeros_like(x)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                potential += func.function(x, y, **kwargs[i])
        return potential

    # @partial(jit, static_argnums=(0, 4))
    def alpha(self, x, y, kwargs, k=None):

        """
        deflection angles
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: deflection angles in units of arcsec
        """
        # x = np.array(x, dtype=float)
        # y = np.array(y, dtype=float)
        if isinstance(k, int):
            return self.func_list[k].derivatives(x, y, **kwargs[k])
        elif self._repeated_profile_mode:
            return self._alpha_repeated(x, y, kwargs, k=k)
        elif self._use_jax_scan:
            return self._alpha_scan(x, y, kwargs, k=k)
        else:
            return self._alpha_loop(x, y, kwargs, k=k)

    def _alpha_repeated(self, x, y, kwargs, k=None):
        if k is not None:
            raise NotImplementedError   # TODO: implement case with k not None
        alpha_func = alpha_static_single(x, y, self.func_list[0].derivatives)
        kwargs_stack = tree_stack(kwargs)
        return jnp.array(jax.vmap(alpha_func)(**kwargs_stack)).sum(axis=1)

    def _alpha_loop(self, x, y, kwargs, k=None):
        bool_list = self._bool_list(k)
        f_x, f_y = jnp.zeros_like(x), jnp.zeros_like(x)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                f_x_i, f_y_i = func.derivatives(x, y, **kwargs[i])
                f_x += f_x_i
                f_y += f_y_i
        return f_x, f_y

    def _alpha_scan(self, x, y, kwargs, k=None):
        """
        Deflection angles using jax.lax.scan, which can lead to faster compilation time 

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: deflection angles in units of arcsec
        """
        if k is not None:
            raise NotImplementedError("Specific k is not yet supported when using jax.lax.scan for deflection angles.")
        # below is a solution for scanning over the profiles based on
        # list of alpha functions with keywords filled in (credits to @krawczyk)
        alpha_list = [
            partial(
                self.func_list[i].derivatives,
                **kwargs[i],
            ) for i in range(self._num_func)
        ]

        # recursive function with keywords filled in
        partial_alpha = partial(
            alpha_static_for_scan,
            x, y,
            alpha_list
        )

        # run recursion function
        f_x, f_y = jnp.zeros_like(x), jnp.zeros_like(y)
        (f_x, f_y), _ = lax.scan(
            partial_alpha,
            (f_x, f_y),
            jnp.arange(self._num_func),
        )
        return f_x, f_y

    def hessian(self, x, y, kwargs, k=None):
        """
        hessian matrix
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: f_xx, f_xy, f_yx, f_yy components
        """
        # x = np.array(x, dtype=float)
        # y = np.array(y, dtype=float)
        if isinstance(k, int):
            f_xx, f_yy, f_xy = self.func_list[k].hessian(x, y, **kwargs[k])
            return f_xx, f_xy, f_xy, f_yy

        bool_list = self._bool_list(k)
        f_xx, f_yy, f_xy = jnp.zeros_like(x), jnp.zeros_like(x), jnp.zeros_like(x)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                f_xx_i, f_yy_i, f_xy_i = func.hessian(x, y, **kwargs[i])
                f_xx += f_xx_i
                f_yy += f_yy_i
                f_xy += f_xy_i
        f_yx = f_xy
        return f_xx, f_xy, f_yx, f_yy

    def kappa(self, x, y, kwargs, k=None):
        """
        lensing convergence k = 1/2 laplacian(phi)

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing convergence
        """
        f_xx, f_xy, f_yx, f_yy = self.hessian(x, y, kwargs, k=k)
        kappa = (f_xx + f_yy) / 2.
        return kappa

    def curl(self, x, y, kwargs, k=None):
        """
        curl computation F_yx - F_xy

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: curl at position (x, y)
        """
        f_xx, f_xy, f_yx, f_yy = self.hessian(x, y, kwargs, k=k)
        # Note the sign change from lenstronomy
        return f_yx - f_xy

    def gamma(self, x, y, kwargs, k=None):
        """
        shear computation
        g1 = 1/2(d^2phi/dx^2 - d^2phi/dy^2)
        g2 = d^2phi/dxdy

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: gamma1, gamma2
        """

        f_xx, f_xy, f_yx, f_yy = self.hessian(x, y, kwargs, k=k)
        gamma1 = (f_xx - f_yy) / 2.
        gamma2 = f_xy
        return gamma1, gamma2

    def inverse_magnification(self, x, y, kwargs, k=None):
        """
        inverse magnification det(A)
        A = 1 - d^2phi/d_ij

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: magnification
        """

        f_xx, f_xy, f_yx, f_yy = self.hessian(x, y, kwargs, k=k)
        det_A = (1 - f_xx) * (1 - f_yy) - f_xy * f_yx
        return det_A
    
    def magnification(self, x, y, kwargs, k=None):
        """
        magnification
        mag = 1/det(A)
        A = 1 - d^2phi/d_ij

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: magnification
        """

        return 1. / self.inverse_magnification(x, y, kwargs, k=k)  # attention, if dividing by zero


class ZeroMassModel(MassModelBase):
    """Dummy mass model that does not have any mass associated with it."""

    def __init__(self, **kwargs):
        self.profile_type_list = [ZeroMass()]
        super().__init__(self.profile_type_list, **kwargs)

    @partial(jit, static_argnums=(0, 4))
    def ray_shooting(self, x, y, kwargs, k=None):
        """
        maps image to source position (inverse deflection)
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: source plane positions corresponding to (x, y) in the image plane
        """
        return x, y

    def fermat_potential(self, x_image, y_image, kwargs_lens, x_source=None, y_source=None, k=None):
        """
        fermat potential (negative sign means earlier arrival time)

        :param x_image: image position
        :param y_image: image position
        :param x_source: source position
        :param y_source: source position
        :param kwargs_lens: list of keyword arguments of lens model parameters matching the lens model classes
        :return: fermat potential in arcsec**2 without geometry term (second part of Eqn 1 in Suyu et al. 2013) as a list
        """
        return jnp.zeros_like(x_image)

    def potential(self, x, y, kwargs, k=None):
        """
        lensing potential
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing potential in units of arcsec^2
        """
        return jnp.zeros_like(x)

    # @partial(jit, static_argnums=(0, 4))
    def alpha(self, x, y, kwargs, k=None):

        """
        deflection angles
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: deflection angles in units of arcsec
        """
        return jnp.array([0.*x, 0.*y])

    def hessian(self, x, y, kwargs, k=None):
        """
        hessian matrix
        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: f_xx, f_xy, f_yx, f_yy components
        """
        return jnp.array([0.*x, 0.*x, 0.*x, 0.*x])

    def kappa(self, x, y, kwargs, k=None):
        """
        lensing convergence k = 1/2 laplacian(phi)

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing convergence
        """
        return jnp.zeros_like(x)

    def curl(self, x, y, kwargs, k=None):
        """
        curl computation F_yx - F_xy

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: curl at position (x, y)
        """
        return jnp.zeros_like(x)

    def gamma(self, x, y, kwargs, k=None):
        """
        shear computation
        g1 = 1/2(d^2phi/dx^2 - d^2phi/dy^2)
        g2 = d^2phi/dxdy

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: gamma1, gamma2
        """
        return jnp.array([0.*x, 0.*x])

    def magnification(self, x, y, kwargs, k=None):
        """
        magnification
        mag = 1/det(A)
        A = 1 - d^2phi/d_ij

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the lens model classes
        :param k: only evaluate the k-th lens model
        :return: magnification
        """
        return jnp.zeros_like(x)

def flatten_func(self):
    children = ()
    aux_data = {}
    return (children, aux_data)

def unflatten_func(aux_data, children):
    return ZeroMassModel()

register_pytree_node(ZeroMassModel, flatten_func, unflatten_func)



# Defines the model of a strong lens
#
# Copyright (c) 2023, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the ImSim module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'

import copy
import numpy as np
import jax.numpy as jnp
from functools import partial
from jax import jit
from jax import random

from herculens.LensImage.Numerics.numerics import Numerics
from herculens.LensImage.lensing_operator import LensingOperator


__all__ = ['LensImage', 'LensImage3D']


class LensImage(object):
    """Generate lensed images from source light, lens mass/light, and point source models."""

    def __init__(self, grid_class, psf_class,
                 noise_class=None, lens_mass_model_class=None,
                 source_model_class=None, lens_light_model_class=None,
                 point_source_model_class=None, 
                 source_arc_mask=None,
                 kwargs_numerics=None,
                 kwargs_lens_equation_solver=None):
        """
        :param grid_class: coordinate system, instance of PixelGrid() from herculens.Coordinates.pixel_grid
        :param psf_class: point spread function, instance of PSF() from herculens.Instrument.psf
        :param noise_class: noise properties, instance of Noise() from herculens.Instrument.noise
        :param lens_mass_model_class: lens mass model, instance of MassModel() from herculens.MassModel.mass_model
        :param source_model_class: extended source light model, instance of LightModel() from herculens.MassModel.mass_model
        :param lens_light_model_class: lens light model, instance of LightModel() from herculens.MassModel.mass_model
        :param point_source_model_class: point source model, instance of PointSourceModel() from herculens.PointSource.point_source
        :param source_arc_mask: 2D boolean array to define the region over which the (pixelated) lensed source is modeled
        :param kwargs_numerics: keyword arguments for various numerical settings (see herculens.Numerics.numerics)
        :param kwargs_lens_equation_solver: TODO
        """
        self.Grid = grid_class
        self.PSF = psf_class
        self.Noise = noise_class

        # Require now that all relevant parameters of the PSF model are provided
        # when it is instantiated (outside this object). This helps avoid JAX
        # tracer errors due to, e.g., the PSF kernel size being computed for
        # the first time inside the jitted model() method, where it would
        # cause the kernel to become an abstract tracer
        #
        # self.PSF.set_pixel_size(self.Grid.pixel_width)

        if lens_mass_model_class is None:
            from herculens.MassModel.mass_model import MassModel
            lens_mass_model_class = MassModel(profile_list=[])
        self.MassModel = lens_mass_model_class
        if self.MassModel.has_pixels:
            pixel_grid = self.Grid.create_model_grid(
                **self.MassModel.pixel_grid_settings)
            self.MassModel.set_pixel_grid(pixel_grid)

        if source_model_class is None:
            from herculens.LightModel.light_model import LightModel
            source_model_class = LightModel(profile_list=[])
        self.SourceModel = source_model_class
        if self.SourceModel.has_pixels:
            pixel_grid = self.Grid.create_model_grid(
                **self.SourceModel.pixel_grid_settings)
            self.SourceModel.set_pixel_grid(pixel_grid, self.Grid.pixel_area)

        if lens_light_model_class is None:
            from herculens.LightModel.light_model import LightModel
            lens_light_model_class = LightModel(profile_list=[])
        self.LensLightModel = lens_light_model_class
        if self.LensLightModel.has_pixels:
            pixel_grid = self.Grid.create_model_grid(
                **self.LensLightModel.pixel_grid_settings)
            self.LensLightModel.set_pixel_grid(
                pixel_grid, self.Grid.pixel_area)

        if point_source_model_class is None:
            from herculens.PointSourceModel.point_source_model import PointSourceModel
            point_source_model_class = PointSourceModel(
                point_source_type_list=[], mass_model=self.MassModel)
        self.PointSourceModel = point_source_model_class
        # self.PointSource.update_lens_model(lens_model_class=lens_model_class)
        # x_center, y_center = self.Data.center
        # self.PointSource.update_search_window(search_window=np.max(self.Data.width), x_center=x_center,
        #                                       y_center=y_center, min_distance=self.Data.pixel_width,
        #                                       only_from_unspecified=True)

        self.source_arc_mask = source_arc_mask
        self._src_adaptive_grid = self.SourceModel.pixel_is_adaptive
        if self._src_adaptive_grid is True and self.source_arc_mask is None:
            raise ValueError("An arc mask for the lensed source must be "
                             "provided with adaptive source grid")
        if self.source_arc_mask is not None:
            self._src_arc_mask_bool = source_arc_mask.astype(bool)

        if kwargs_numerics is None:
            kwargs_numerics = {}
        self.kwargs_numerics = kwargs_numerics
        self.ImageNumerics = Numerics(
            pixel_grid=self.Grid, psf=self.PSF, **self.kwargs_numerics)

        self.kwargs_lens_equation_solver = kwargs_lens_equation_solver

    # WIP
    def set_static_model_grid(self):
        ra_grid_img, dec_grid_img = self.ImageNumerics.coordinates_evaluate
        for j in range(self.MassModel._num_func):
            profile = self.MassModel.func_list[j]
            if hasattr(profile, 'set_eval_coord_grid'):
                profile.set_eval_coord_grid(
                    ra_grid_img.reshape(*self.ImageNumerics._grid.num_grid_points_axes),
                    dec_grid_img.reshape(*self.ImageNumerics._grid.num_grid_points_axes),
                )
                print(f"Set static coordinate grid for profile {j}")

    def source_surface_brightness(self, kwargs_source, kwargs_lens=None,
                                  unconvolved=False, supersampled=False,
                                  de_lensed=False, k=None, k_lens=None,
                                  adapted_pixels_coords=None, 
                                  return_pixels_coords=False,
                                  return_as_list=False):
        """

        computes the source surface brightness distribution

        :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles.
        When using an adaptive source pixel grid, kwargs_lens is required even if de_lensed=True.
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness profile, otherwise the lensed.
        :param k: list of bool or list of int to select which source profiles to include
        :param k_lens: list of bool or list of int to select which lens mass profiles to include
        :param adapted_pixels_coords: tuple of 1d arrays, x and y coordinates of the pixelated light profile
        :param return_pixels_coords: if True, returns the adapted pixel coordinates
        :param return_as_list: if True, returns the source surface brightness as a list of arrays
        :return: 2d array of surface brightness pixels
        """
        if len(self.SourceModel.profile_type_list) == 0:
            source_light = jnp.zeros(self.Grid.num_pixel_axes)
            if return_pixels_coords:
                return source_light, None
            return source_light
        x_grid_img, y_grid_img = self.ImageNumerics.coordinates_evaluate
        source_light, adapted_pixels_coords = self.eval_source_surface_brightness(
            x_grid_img, y_grid_img,
            kwargs_source, kwargs_lens=kwargs_lens,
            k=k, k_lens=k_lens, de_lensed=de_lensed,
            adapted_pixels_coords=adapted_pixels_coords,
            return_pixels_coords=True,
            return_as_list=return_as_list
        )
        if not supersampled:
            source_light = self.ImageNumerics.re_size_convolve(
                source_light, unconvolved=unconvolved, input_as_list=return_as_list,
            )
        if return_pixels_coords:
            return source_light, adapted_pixels_coords
        return source_light
    
    def eval_source_surface_brightness(self, x, y, kwargs_source, kwargs_lens=None, 
                                       k=None, k_lens=None, de_lensed=False,
                                       adapted_pixels_coords=None, 
                                       return_pixels_coords=False,
                                       return_as_list=False):
        if self._src_adaptive_grid:
            if adapted_pixels_coords is None:
                pixels_x_coord, pixels_y_coord, _ = self.adapt_source_coordinates(kwargs_lens, k_lens=k_lens)
            else:
                pixels_x_coord, pixels_y_coord = adapted_pixels_coords
        else:
            pixels_x_coord, pixels_y_coord = None, None
        if de_lensed is True:
            source_light = self.SourceModel.surface_brightness(x, y, kwargs_source, k=k,
                                                               pixels_x_coord=pixels_x_coord, pixels_y_coord=pixels_y_coord,
                                                               return_as_list=return_as_list)
        else:
            x_grid_src, y_grid_src = self.MassModel.ray_shooting(x, y, kwargs_lens, k=k_lens)
            source_light = self.SourceModel.surface_brightness(x_grid_src, y_grid_src, kwargs_source, k=k,
                                                               pixels_x_coord=pixels_x_coord, pixels_y_coord=pixels_y_coord,
                                                               return_as_list=return_as_list)
        if return_pixels_coords:
            return source_light, (pixels_x_coord, pixels_y_coord)
        return source_light

    def lens_surface_brightness(self, kwargs_lens_light, unconvolved=False,
                                supersampled=False, k=None):
        """

        computes the lens surface brightness distribution

        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :param unconvolved: if True, returns unconvolved surface brightness (perfect seeing), otherwise convolved with PSF kernel
        :param k: list of bool or list of int to select which model profiles to include
        :return: 2d array of surface brightness pixels
        """
        x_grid_img, y_grid_img = self.ImageNumerics.coordinates_evaluate
        lens_light = self.LensLightModel.surface_brightness(x_grid_img, y_grid_img, kwargs_lens_light, k=k)
        if not supersampled:
            lens_light = self.ImageNumerics.re_size_convolve(
                lens_light, unconvolved=unconvolved)
        return lens_light

    def point_source_image(self, kwargs_point_source, kwargs_lens,
                           kwargs_solver, k=None):
        """Compute PSF-convolved point sources rendered on the image plane.

        :param kwargs_point_source: list of keyword arguments corresponding to the point sources
        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param k: list of bool or list of int to select which point sources to include
        :return: 2d array at the image plane resolution of all multiple images of the point sources
        """
        result = jnp.zeros((self.Grid.num_pixel_axes))
        if self.PointSourceModel is None:
            return result
        theta_x, theta_y, amplitude = self.PointSourceModel.get_multiple_images(
            kwargs_point_source, kwargs_lens=kwargs_lens, kwargs_solver=kwargs_solver, 
            k=k, with_amplitude=True, zero_amp_duplicates=True
        )
        for i in range(len(theta_x)):
            result += self.ImageNumerics.render_point_sources(
                theta_x[i], theta_y[i], amplitude[i]
            )
        return result

    @partial(jit, static_argnums=(0, 5, 6, 7, 8, 9, 10, 11, 12, 13, 15))
    def model(self, kwargs_lens=None, kwargs_source=None, kwargs_lens_light=None,
              kwargs_point_source=None, unconvolved=False, supersampled=False,
              source_add=True, lens_light_add=True, point_source_add=True,
              k_lens=None, k_source=None, k_lens_light=None, k_point_source=None,
              adapted_source_pixels_coords=None, return_source_pixels_coords=False):
        """
        Create the 2D model image from parameter values.
        Note: due to JIT compilation, the first call to this method will be slower.

        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :param kwargs_point_source: keyword arguments corresponding to "other" parameters, such as external shear and
                                    point source image positions
        :param unconvolved: if True: returns the unconvolved light distribution (prefect seeing)
        :param supersampled: if True, returns the model on the higher resolution grid (WARNING: no convolution nor normalization is performed in this case!)
        :param source_add: if True, compute source, otherwise without
        :param lens_light_add: if True, compute lens light, otherwise without
        :param point_source_add: if True, compute point source multiple images, otherwise without
        :param k_lens: list of bool or list of int to select which lens mass profiles to include
        :param k_source: list of bool or list of int to select which source profiles to include
        :param k_lens_light: list of bool or list of int to select which lens light profiles to include
        :param k_point_source: list of bool or list of int to select which point sources to include
        :return: 2d array of surface brightness pixels of the simulation
        """
        # TODO: simplify treatment of convolution, downsampling and re-sizing
        model = jnp.zeros((self.Grid.num_pixel_axes))
        if supersampled:
            model = jnp.zeros((self.ImageNumerics.grid_class.num_grid_points,))
        if source_add is True:
            if self.source_arc_mask is not None:
                return_as_list = True  # this is to apply the arc mask only to the pixelated source profile
            else:
                return_as_list = False
            source_model, adapted_source_pixels_coords = self.source_surface_brightness(
                kwargs_source, kwargs_lens, unconvolved=unconvolved, 
                supersampled=supersampled, k=k_source, k_lens=k_lens,
                adapted_pixels_coords=adapted_source_pixels_coords, 
                return_pixels_coords=True,
                return_as_list=return_as_list)
            if return_as_list:  # i.e. if self.source_arc_mask is not None
                source_model[self.SourceModel.pixelated_index] *= self.source_arc_mask
                source_model = jnp.sum(jnp.array(source_model), axis=0)
            model += source_model
        if lens_light_add is True:
            model += self.lens_surface_brightness(kwargs_lens_light, 
                                                  unconvolved=unconvolved, supersampled=supersampled,
                                                  k=k_lens_light)
        if point_source_add:
            model += self.point_source_image(kwargs_point_source, kwargs_lens,
                                             kwargs_solver=self.kwargs_lens_equation_solver,
                                             k=k_point_source)
        if return_source_pixels_coords:
            return model, adapted_source_pixels_coords
        return model

    def simulation(self, add_poisson_noise=True, add_background_noise=True,
                   compute_true_noise_map=True, prng_key=random.PRNGKey(18),
                   **model_kwargs):
        """
        same as model() but with noise added

        :param compute_true_noise_map: if True (default), define the noise map (diagonal covariance matrix)
        to be the 'true' one, i.e. based on the noiseless model image.
        :param prng_key: instance of jax.random.PRNGKey to generate the noise realization.
        The default is an arbitrary value from seed 18, so it is the user task to change it for different realizations.
        """
        if self.Noise is None:
            raise ValueError(
                "Impossible to generate noise realisation because no noise class has been set")
        model = self.model(**model_kwargs)
        noise = self.Noise.realisation(
            model, prng_key, 
            add_poisson_model=add_poisson_noise, 
            add_background=add_background_noise)
        simu = model + noise
        self.Noise.set_data(simu)
        if compute_true_noise_map is True:
            self.Noise.compute_noise_map_from_model(model)
        return simu
    
    def C_D_model(self, model, **kwargs_noise):
        return self.Noise.C_D_model(model, **kwargs_noise)

    def normalized_residuals(self, data, model, kwargs_noise=None, mask=None):
        """
        compute the map of normalized residuals,
        given the data and the model image
        """
        if kwargs_noise is None:
            kwargs_noise = {}
        if mask is None:
            mask = np.ones(self.Grid.num_pixel_axes)
        noise_var = self.C_D_model(model, **kwargs_noise)
        noise = np.sqrt(noise_var)
        norm_res_model = (data - model) / noise * mask
        norm_res_tot = norm_res_model
        if mask is not None:
            # outside the mask just add pure data
            norm_res_tot += (data / noise) * (1. - mask)
        # make sure there is no NaN or infinite values
        norm_res_model = np.where(np.isfinite(norm_res_model), norm_res_model, 0.)
        norm_res_tot = np.where(np.isfinite(norm_res_tot), norm_res_tot, 0.)
        return norm_res_model, norm_res_tot

    def reduced_chi2(self, data, model, kwargs_noise=None, mask=None):
        """
        compute the reduced chi2 of the data given the model
        """
        if mask is None:
            mask = np.ones(self.Grid.num_pixel_axes)
        norm_res, _ = self.normalized_residuals(data, model, kwargs_noise=kwargs_noise, mask=mask)
        num_data_points = np.sum(mask)
        return np.sum(norm_res**2) / num_data_points
    
    def adapt_source_coordinates(self, kwargs_lens, k_lens=None, return_plt_extent=False):
        """
        Compute new source coordinates based on ray-traced arc-mask
        return_plt_extent=True returns the extent of the grid for imshow
        """
        # TODO: compute a mask on the adapted source plane
        npix_src, npix_src_y = self.SourceModel.pixel_grid.num_pixel_axes
        if npix_src_y != npix_src:
            raise ValueError("Adaptive source plane grid only works with square grids")
        if self.Grid.x_is_inverted or self.Grid.y_is_inverted:
            # TODO: implement this
            raise NotImplementedError("invert x and y not yet supported for adaptive source grid")
        # get data coordinates
        x_grid_img, y_grid_img = self.Grid.pixel_coordinates
        # ray-shoot to source plane only coordinates within the source arc mask
        x_grid_src, y_grid_src = self.MassModel.ray_shooting(x_grid_img, #[self._src_arc_mask_bool], 
                                                             y_grid_img, #[self._src_arc_mask_bool], 
                                                             kwargs_lens, k=k_lens)
        x_grid_src = x_grid_src[self._src_arc_mask_bool]
        y_grid_src = y_grid_src[self._src_arc_mask_bool]
        # create grid encompassed by ray-shot coordinates
        x_left, x_right = x_grid_src.min(), x_grid_src.max()
        y_bottom, y_top = y_grid_src.min(), y_grid_src.max()
        # center of the region
        cx = 0.5 * (x_left + x_right)
        cy = 0.5 * (y_bottom + y_top)
        # get the width and height
        width  = jnp.abs(x_left - x_right)
        height = jnp.abs(y_bottom - y_top)
        # choose the largest of the two to end up with a square region
        half_size = jnp.maximum(height, width) / 2.
        # recompute the new boundaries
        x_left = cx - half_size
        x_right = cx + half_size
        y_bottom = cy - half_size
        y_top = cy + half_size
        # print( 0.5*(x_left + x_right), cx )
        # print( jnp.abs(x_left - x_right), size )
        # print( 0.5*(y_bottom + y_top), cy )
        # print( jnp.abs(y_bottom - y_top), size )
        x_adapt = jnp.linspace(x_left, x_right, npix_src)
        y_adapt = jnp.linspace(y_bottom, y_top, npix_src)
        if return_plt_extent is True:
            pix_scl_x = jnp.abs(x_adapt[0]-x_adapt[1])
            pix_scl_y = jnp.abs(y_adapt[0]-y_adapt[1])
            half_pix_scl = jnp.sqrt(pix_scl_x*pix_scl_y) / 2.
            extent_adapt = [
                x_adapt[0]-half_pix_scl, x_adapt[-1]+half_pix_scl, 
                y_adapt[0]-half_pix_scl, y_adapt[-1]+half_pix_scl
            ]
        else:
            extent_adapt = [
                x_adapt[0], x_adapt[-1], 
                y_adapt[0], y_adapt[-1]
            ]
        return x_adapt, y_adapt, extent_adapt
    
    def get_source_coordinates(self, kwargs_lens, k_lens=None, return_plt_extent=False):
        if not self._src_adaptive_grid:
            x_grid, y_grid = self.SourceModel.pixel_grid.pixel_coordinates
            if return_plt_extent is True:
                extent = self.SourceModel.pixel_grid.plt_extent
            else:
                extent = self.SourceModel.pixel_grid.extent
        else:
            x_coord, y_coord, extent = self.adapt_source_coordinates(kwargs_lens, k_lens=k_lens,
                                                                     return_plt_extent=return_plt_extent)
            x_grid, y_grid = np.meshgrid(x_coord, y_coord)
        return x_grid, y_grid, extent
        
    def get_lensing_operator(self, kwargs_lens=None, update=False, arc_mask=None):
        if self.SourceModel.pixel_grid is None:
            raise ValueError("The lensing operator is only defined for source "
                            "models associated to a grid of pixels")
        if not hasattr(self, '_lensing_op'):
            self._lensing_op = LensingOperator(self.MassModel,
                                            self.Grid, # TODO: should be the model grid (from Numerics) at some point
                                            self.SourceModel.pixel_grid,
                                            arc_mask=arc_mask)
        if update is True or self._lensing_op.get_lens_mapping() is None:
            self._lensing_op.compute_mapping(kwargs_lens)
        return self._lensing_op
    


class LensImage3D(object):
    """
    Handy class to jointly handle a set of multiple LensImage models.
    Typical use-cases are the joint modelling of multiple bands or multiple exposures.

    NOTE: Special care must be taken when using LensImage instances that represent models at 
    different resolutions (pixel sizes, cutout size, etc.) in conjunction with
    pixelated source models. In case you use a 3D pixelated source model 
    (such as CorrelatedField model) on an adaptive grid, you may want to set 
    `join_source_coords=True` to ensure that the source model gets evaluated on the same grid.
    """

    def __init__(self, lens_image_list, mode='stack', reference_lens_image=0, join_source_coords=True):
        """
        Initialize a LensImage object.

        Parameters
        ----------
        lens_image_list : list
            List of LensImages instances to combine.
        mode : str, optional
            Either 'list' or 'stack', defining the output format of the main methods, such as .model(). By default 'stack'.
        reference_lens_image : int, optional
            Index of the reference lens image. By default 0.
        join_source_coords : bool, optional
            If set to True, the adaptive source pixelated grid of the 
            reference LensImage instance (if any) will be used to define 
            all source pixelated grids. By default True.

        Raises
        ------
        ValueError
            If `lens_image_list` contains only one LensImage instance.
        ValueError
            If `mode` is not 'stack' or 'list'.
        ValueError
            If the reference LensImage instance does not have an adaptive source grid
            and joint_source_coords was set to True.

        Notes
        -----
        The `__init__` method initializes a LensImage object with the given parameters. It sets the `mode` attribute to the specified mode, the `num_bands` attribute to the number of lens images in the list, and the `lens_images` attribute to the provided list of lens images. It also checks if the `mode` is 'stack' and verifies that all lens images have the same shape.

        Examples
        --------
        >>> lens_image_list = [lens_image1, lens_image2, lens_image3]
        >>> lens = LensImage3D(lens_image_list, mode='stack', join_source_coords=True, reference_lens_image=0)
        """
        if len(lens_image_list) == 1:
            raise ValueError("LensImage3D should be used with more than one band.")
        if mode not in ('stack', 'list'):
            raise ValueError("mode should be either 'stack' or 'list'.")
        self.mode = mode
        self.num_bands = len(lens_image_list)
        if self.mode == 'stack':
            for i in range(1, self.num_bands):
                if lens_image_list[i].Grid.num_pixel_axes != lens_image_list[i-1].Grid.num_pixel_axes:
                    raise ValueError("In each band, all models should have the same shape.")
            self.nx, self.ny = lens_image_list[0].Grid.num_pixel_axes
            self.nw = self.num_bands
        else:
            self.nx, self.ny, self.nw = None, None, None
        self.lens_images = lens_image_list
        self._idx_ref = reference_lens_image
        if any([lens_image.SourceModel.pixel_is_adaptive for lens_image in lens_image_list]):
            self._join_source_coords = join_source_coords
        else:
            self._join_source_coords = False
        if self._join_source_coords and not self.ref_lens_image.SourceModel.pixel_is_adaptive:
            raise ValueError("The reference LensImage instance must have an "
                             "adaptive source grid to join source coordinates.")

    @property
    def ref_lens_image(self):
        """Reference LensImage instance (used e.g. for joining source pixelated grids)."""
        return self.lens_images[self._idx_ref]

    def source_surface_brightness(self, kwargs_source_list, k_list=None, **kwargs_single_band):
        """

        computes the source surface brightness distribution

        :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness profile, otherwise the lensed.
        :param k: list of bool or list of int to select which source profiles to include
        :param k_lens: list of bool or list of int to select which lens mass profiles to include
        :return: 2d array of surface brightness pixels
        """
        # TODO: add support for join_source_coords
        if k_list is None:
            k_list = [None]*self.num_bands
        source_light_list = []
        for i in range(self.num_bands):
            source_light = self.lens_images[i].source_surface_brightness(
                kwargs_source_list[i], k=k_list[i], **kwargs_single_band,
            )
            source_light_list.append(source_light)
        return source_light_list

    def lens_surface_brightness(self, kwargs_lens_light_list, k_list=None, **kwargs_single_band):
        """

        computes the lens surface brightness distribution

        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :param unconvolved: if True, returns unconvolved surface brightness (perfect seeing), otherwise convolved with PSF kernel
        :param k: list of bool or list of int to select which model profiles to include
        :return: 2d array of surface brightness pixels
        """
        if k_list is None:
            k_list = [None]*self.num_bands
        lens_light_list = []
        for i in range(self.num_bands):
            lens_light = self.lens_images[i].lens_surface_brightness(
                kwargs_lens_light_list[i], k=k_list[i], **kwargs_single_band,
            )
            lens_light_list.append(lens_light)
        return lens_light_list

    def point_source_image(self, kwargs_point_source_list, k_list=None, **kwargs_single_band):
        """Compute PSF-convolved point sources rendered on the image plane.

        :param kwargs_point_source: list of keyword arguments corresponding to the point sources
        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param k: list of bool or list of int to select which point sources to include
        :return: 2d array at the image plane resolution of all multiple images of the point sources
        """
        if k_list is None:
            k_list = [None]*self.num_bands
        point_source_list = []
        for i in range(self.num_bands):
            point_source = self.lens_images[i].point_source_image(
                kwargs_point_source_list[i], k=k_list[i], **kwargs_single_band,
            )
            point_source_list.append(point_source)
        return point_source_list

    @partial(jit, static_argnums=(0, 5, 6, 7, 8, 9, 10, 11, 12, 13))
    def model(self, kwargs_lens=None, kwargs_source_list=None, kwargs_lens_light_list=None,
              kwargs_point_source_list=None, unconvolved=False, supersampled=False,
              source_add=True, lens_light_add=True, point_source_add=True,
              k_lens=None, k_source_list=None, k_lens_light_list=None, k_point_source_list=None):
        """
        Create the 2D model image from parameter values.
        Note: due to JIT compilation, the first call to this method will be slower.

        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :param kwargs_point_source: keyword arguments corresponding to "other" parameters, such as external shear and
                                    point source image positions
        :param unconvolved: if True: returns the unconvolved light distribution (prefect seeing)
        :param supersampled: if True, returns the model on the higher resolution grid (WARNING: no convolution nor normalization is performed in this case!)
        :param source_add: if True, compute source, otherwise without
        :param lens_light_add: if True, compute lens light, otherwise without
        :param point_source_add: if True, compute point source multiple images, otherwise without
        :param k_lens: list of bool or list of int to select which lens mass profiles to include
        :param k_source: list of bool or list of int to select which source profiles to include
        :param k_lens_light: list of bool or list of int to select which lens light profiles to include
        :param k_point_source: list of bool or list of int to select which point sources to include
        :return: 2d array of surface brightness pixels of the simulation
        """
        kwargs_source_list = self._none_kwargs(kwargs_source_list)
        kwargs_lens_light_list = self._none_kwargs(kwargs_lens_light_list)
        kwargs_point_source_list = self._none_kwargs(kwargs_point_source_list)
        k_source_list = self._none_kwargs(k_source_list)
        k_lens_light_list = self._none_kwargs(k_lens_light_list)
        k_point_source_list = self._none_kwargs(k_point_source_list)
        # if we join the source coordinates, we first need to get them from the reference band
        if self._join_source_coords:
            src_x_coord, src_y_coord, _ = self.ref_lens_image.adapt_source_coordinates(
                kwargs_lens, 
                k_lens=None,  # we consider all mass profiles for this
                return_plt_extent=False,
            )
            source_coords = (src_x_coord, src_y_coord)
        else:
            source_coords = None
        model_multi_band = []
        for i in range(self.num_bands):
            model_sgl_band = self.lens_images[i].model(
                kwargs_lens=kwargs_lens,
                kwargs_source=kwargs_source_list[i],
                kwargs_lens_light=kwargs_lens_light_list[i],
                kwargs_point_source=kwargs_point_source_list[i],
                unconvolved=unconvolved, supersampled=supersampled,
                source_add=source_add, lens_light_add=lens_light_add, 
                point_source_add=point_source_add,
                k_lens=k_lens, k_lens_light=k_lens_light_list[i], k_source=k_source_list[i], 
                k_point_source=k_point_source_list[i],
                adapted_source_pixels_coords=source_coords,
            )
            model_multi_band.append(model_sgl_band)
        if self.mode == 'stack':
            return jnp.stack(model_multi_band, axis=0)
        return model_multi_band

    def simulation(self, *args, **kwargs):
        raise NotImplementedError("The .simulation() method is not supported for 3D models. "
                                  "Please use the standard LensImage class.")
    
    def C_D_model(self, model):
        c_d_multi_band = []
        for i in range(self.num_bands):
            c_d_slg_band = self.lens_images[i].C_D_model(model[i])
            c_d_multi_band.append(c_d_slg_band)
        if self.mode == 'stack':
            return jnp.array(c_d_multi_band)
        return c_d_multi_band

    def normalized_residuals(self, data, model, mask=None):
        """
        compute the map of normalized residuals,
        given the data and the model image
        """
        # TODO: add support for mode='list'
        if self.mode == 'list':
            raise NotImplementedError(
                "The normalized_residuals() method is not supported for 3D models returned as lists. "
                "Please use the individual LensImage instances."
            )
        if mask is None:
            mask = np.ones_like(data)
        noise_var = self.C_D_model(model)
        noise = np.sqrt(noise_var)
        norm_res_model = (data - model) / noise * mask
        norm_res_tot = norm_res_model
        if mask is not None:
            # outside the mask just add pure data
            norm_res_tot += (data / noise) * (1. - mask)
        return norm_res_model, norm_res_tot

    def reduced_chi2(self, data, model, mask=None):
        """
        compute the reduced chi2 of the data given the model
        """
        # TODO: add support for mode='list'
        if self.mode == 'list':
            raise NotImplementedError(
                "The reduced_chi2() method is not supported for 3D models returned as lists. "
                "Please use the individual LensImage instances."
            )
        if mask is None:
            mask = np.ones_like(data)
        norm_res, _ = self.normalized_residuals(data, model, mask=mask)
        num_data_points = np.sum(mask)
        return np.sum(norm_res**2) / num_data_points
    
    def _none_kwargs(self, kwargs):
        return kwargs if kwargs is not None else [None]*self.num_bands
    


# Handles coordinate grid on which ray-tracing and convolution are performed
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the ImSim.Numerics module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import jax.numpy as np
from herculens.Util import util
from herculens.Util import image_util
from herculens.Coordinates.coord_transforms import Coordinates1D


__all__ = ['RegularGrid']


class RegularGrid(Coordinates1D):
    """
    manages a super-sampled grid on the partial image
    """
    def __init__(self, nx, ny, transform_pix2angle, ra_at_xy_0, dec_at_xy_0, supersampling_factor=1):
        """

        :param nx: number of pixels in x-axis
        :param ny: number of pixels in y-axis
        :param transform_pix2angle: 2x2 matrix, mapping of pixel to coordinate
        :param ra_at_xy_0: ra coordinate at pixel (0,0)
        :param dec_at_xy_0: dec coordinate at pixel (0,0)
        :param supersampling_indexes: bool array of shape nx x ny, corresponding to pixels being super_sampled
        :param supersampling_factor: int, factor (per axis) of super-sampling
        :param flux_evaluate_indexes: bool array of shape nx x ny, corresponding to pixels being evaluated
        (for both low and high res). Default is None, replaced by setting all pixels to being evaluated.
        """
        super(RegularGrid, self).__init__(transform_pix2angle, ra_at_xy_0, dec_at_xy_0)
        self._supersampling_factor = supersampling_factor
        self._nx = nx
        self._ny = ny
        self._x_grid, self._y_grid = self.coordinate_grid(nx, ny)
        x_grid_sub, y_grid_sub = util.subgrid_from_coordinate_transform(self._nx, self._nx,
                                                                   transform_pix2angle, ra_at_xy_0, dec_at_xy_0,
                                                                   subgrid_res=self._supersampling_factor)
        self._ra_subgrid = x_grid_sub
        self._dec_subgrid = y_grid_sub

    @property
    def coordinates_evaluate(self):
        """

        :return: 1d array of all coordinates being evaluated to perform the image computation
        """
        return self._ra_subgrid, self._dec_subgrid

    @property
    def grid_points_spacing(self):
        """
        effective spacing between coordinate points, after supersampling
        :return: sqrt(pixel_area)/supersampling_factor
        """
        return self.pixel_width / self._supersampling_factor

    @property
    def num_grid_points_axes(self):
        """
        effective number of points along each axes, after supersampling
        :return: number of pixels per axis, nx*supersampling_factor ny*supersampling_factor
        """
        return self._nx * self._supersampling_factor, self._ny * self._supersampling_factor

    @property
    def num_grid_points(self):
        """
        effective number of points along each axes, after supersampling
        :return: number of pixels per axis, nx*supersampling_factor ny*supersampling_factor
        """
        return self._nx * self._supersampling_factor * self._ny * self._supersampling_factor

    @property
    def supersampling_factor(self):
        """
        :return: factor (per axis) of super-sampling relative to a pixel
        """
        return self._supersampling_factor

    def flux_array2image_low_high(self, flux_array, **kwargs):
        """

        :param flux_array: 1d array of low and high resolution flux values corresponding to the coordinates_evaluate order
        :return: 2d array, 2d array, corresponding to (partial) images in low and high resolution (to be convolved)
        """
        image = self._array2image(flux_array)
        if self._supersampling_factor > 1:
            image_high_res = image
            image_low_res = image_util.re_size(image, self._supersampling_factor)
        else:
            image_high_res = None
            image_low_res = image
        return image_low_res, image_high_res

    def _array2image(self, array):
        """
        maps a 1d array into a (nx, ny) 2d grid with array populating the idex_mask indices

        :param array: 1d array
        :param idex_mask: 1d array of length nx*ny
        :param nx: x-axis of 2d grid
        :param ny: y-axis of 2d grid
        :return:
        """
        nx, ny = self._nx * self._supersampling_factor, self._ny * self._supersampling_factor
        return util.array2image(array, nx, ny)



# Defines regularization choices
# 
# Copyright (c) 2023, herculens developers and contributors


__author__ = 'aymgal'


import jax.numpy as jnp

from herculens.RegulModel.Methods.base import BaseRegularization
from herculens.RegulModel import regul_util


__all__ = [
    'SparsityStarlet',
    'SparsityBLWavelet',
]


class BaseSparsityWaveletAnalysis(BaseRegularization):

    """
    Base class for regularization strategies imposing a sparsity constraint
    in transformed space (a.k.a. analysis framework)
    """

    def __init__(self, model_type, profile_index, mass_form=None):
        super().__init__(model_type, profile_index, mass_form=mass_form)
        self.transform = None
        self.weights = None
        self._transform_name = None
        self._second_gen = None

    def initialize(self, lens_image, kwargs_params, **kwargs_weights):
        if self.model_type in ('source', 'lens_light'):
            fn = regul_util.data_noise_to_wavelet_light
        elif self.model_type == 'lens_mass' and self._mass_form == 'potential':
            fn = regul_util.data_noise_to_wavelet_potential
        else:
            raise ValueError(f"Combination of {self.model_type} with "
                             f"lens mass '{self._mass_form}' is not supported")
        weights_list, transform_list = fn(lens_image, kwargs_params,
                                          model_type=self.model_type,
                                          wavelet_type_list=[self._transform_name],  # TODO: improve interface
                                          starlet_second_gen=self._second_gen,
                                          **kwargs_weights)
        self.transform = transform_list[0]
        self.weights = weights_list[0]


class SparsityStarlet(BaseSparsityWaveletAnalysis):

    param_names = ['lambda_0', 'lambda_1']
    lower_limit_default = {'lambda_0': 0, 'lambda_1': 0}
    upper_limit_default = {'lambda_0': 1e8, 'lambda_1': 1e8}
    fixed_default = {key: True for key in param_names}

    def __init__(self, model_type, profile_index, mass_form=None, second_gen=False):
        super().__init__(model_type, profile_index, mass_form=mass_form)
        self._transform_name = 'starlet'
        self._second_gen = second_gen

    def log_prob(self, kwargs_params, lambda_0=0, lambda_1=0):
        # get the pixels to be regularized
        pixels = self.get_pixel_params(kwargs_params)
        # pixels in transformed space
        coeffs = self.transform.decompose(pixels)
        # regularization weights
        W = self.weights
        # first scale (i.e. high frequencies)
        l1_weighted_coeffs0 = jnp.sum(jnp.abs(W[0] * coeffs[0]))
        # other scales (except coarsest)
        l1_weighted_coeffs1 = jnp.sum(jnp.abs(W[1:-1] * coeffs[1:-1]))
        # sum the two terms and minus sign to get log-probability
        log_prob = - lambda_0 * l1_weighted_coeffs0 - lambda_1 * l1_weighted_coeffs1
        return log_prob


class SparsityBLWavelet(BaseSparsityWaveletAnalysis):

    param_names = ['lambda_0']
    lower_limit_default = {'lambda_0': 0}
    upper_limit_default = {'lambda_0': 1e8}
    fixed_default = {key: True for key in param_names}

    def __init__(self, model_type, profile_index, mass_form=None):
        super().__init__(model_type, profile_index, mass_form=mass_form)
        self._transform_name = 'battle-lemarie-3'

    def log_prob(self, kwargs_params, lambda_0=0):
        # get the pixels to be regularized
        pixels = self.get_pixel_params(kwargs_params)
        # pixels in transformed space
        coeffs = self.transform.decompose(pixels)
        # regularization weights
        W = self.weights
        # only first scale (i.e. high frequencies)
        l1_weighted_coeffs0 = jnp.sum(jnp.abs(W[0] * coeffs[0]))
        log_prob = - lambda_0 * l1_weighted_coeffs0
        return log_prob



# Utility functions
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Util module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'dangilman', 'aymgal'

import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy.ndimage import interpolation as interp
from jax import random
import jax.numpy as jnp

from objax import functional
from objax.constants import ConvPadding
# from herculens.Util.jax_util import BilinearInterpolator


def add_layer2image(grid2d, x_pos, y_pos, kernel, order=1):
    """
    adds a kernel on the grid2d image at position x_pos, y_pos with an interpolated subgrid pixel shift of order=order
    :param grid2d: 2d pixel grid (i.e. image)
    :param x_pos: x-position center (pixel coordinate) of the layer to be added
    :param y_pos: y-position center (pixel coordinate) of the layer to be added
    :param kernel: the layer to be added to the image
    :param order: interpolation order for sub-pixel shift of the kernel to be added
    :return: image with added layer, cut to original size
    """

    x_int = int(round(x_pos))
    y_int = int(round(y_pos))
    shift_x = x_int - x_pos
    shift_y = y_int - y_pos
    kernel_shifted = interp.shift(kernel, [-shift_y, -shift_x], order=order)
    return add_layer2image_int(grid2d, x_int, y_int, kernel_shifted)

def add_layer2image_int(grid2d, x_pos, y_pos, kernel):
    """
    adds a kernel on the grid2d image at position x_pos, y_pos at integer positions of pixel
    :param grid2d: 2d pixel grid (i.e. image)
    :param x_pos: x-position center (pixel coordinate) of the layer to be added
    :param y_pos: y-position center (pixel coordinate) of the layer to be added
    :param kernel: the layer to be added to the image
    :return: image with added layer
    """
    nx, ny = np.shape(kernel)
    if nx % 2 == 0:
        raise ValueError("kernel needs odd numbers of pixels")

    num_x, num_y = np.shape(grid2d)
    x_int = int(round(x_pos))
    y_int = int(round(y_pos))

    k_x, k_y = np.shape(kernel)
    k_l2_x = int((k_x - 1) / 2)
    k_l2_y = int((k_y - 1) / 2)

    min_x = np.maximum(0, x_int-k_l2_x)
    min_y = np.maximum(0, y_int-k_l2_y)
    max_x = np.minimum(num_x, x_int+k_l2_x + 1)
    max_y = np.minimum(num_y, y_int+k_l2_y + 1)

    min_xk = np.maximum(0, -x_int + k_l2_x)
    min_yk = np.maximum(0, -y_int + k_l2_y)
    max_xk = np.minimum(k_x, -x_int + k_l2_x + num_x)
    max_yk = np.minimum(k_y, -y_int + k_l2_y + num_y)
    if min_x >= max_x or min_y >= max_y or min_xk >= max_xk or min_yk >= max_yk or (max_x-min_x != max_xk-min_xk) or (max_y-min_y != max_yk-min_yk):
        return grid2d
    kernel_re_sized = kernel[min_yk:max_yk, min_xk:max_xk]
    new = grid2d.copy()
    new[min_y:max_y, min_x:max_x] += kernel_re_sized
    return new

def add_background(image, sigma_bkd, seed):
    """
    adds background noise to image
    :param image: pixel values of image
    :param sigma_bkd: background noise (sigma)
    :return: a realisation of Gaussian noise of the same size as image
    """
    background = random.normal(seed, shape=np.shape(image)) * sigma_bkd
    # without JAX:
    # nx, ny = 
    # background = np.random.randn(*image.shape) * sigma_bkd
    return background

def add_poisson(image, exp_time, seed):
    """
    adds a poison (or Gaussian) distributed noise with mean given by surface brightness
    :param image: pixel values (photon counts per unit exposure time)
    :param exp_time: exposure time
    :return: Poisson noise realization of input image
    """
    """
    adds a poison (or Gaussian) distributed noise with mean given by surface brightness
    """
    sigma = jnp.sqrt(jnp.abs(image) / exp_time) # Gaussian approximation for Poisson distribution, normalized to exposure time
    poisson = random.normal(seed, shape=jnp.shape(image)) * sigma
    # without JAX:
    # sigma = np.sqrt(np.abs(image) / exp_time) # Gaussian approximation for Poisson distribution, normalized to exposure time
    # poisson = np.random.randn(*image.shape) * sigma
    return poisson

def cut_edges(image, numPix):
    """
    cuts out the edges of a 2d image and returns re-sized image to numPix
    center is well defined for odd pixel sizes.
    :param image: 2d numpy array
    :param numPix: square size of cut out image
    :return: cutout image with size numPix
    """
    nx, ny = image.shape
    if nx < numPix or ny < numPix:
        raise ValueError('image can not be resized, in routine cut_edges with image shape (%s %s) '
                         'and desired new shape (%s %s)' % (nx, ny, numPix, numPix))
    if (nx % 2 == 0 and ny % 2 == 1) or (nx % 2 == 1 and ny % 2 == 0):
        raise ValueError('image with odd and even axis (%s %s) not supported for re-sizeing' % (nx, ny))
    if (nx % 2 == 0 and numPix % 2 == 1) or (nx % 2 == 1 and numPix % 2 == 0):
        raise ValueError('image can only be re-sized from even to even or odd to odd number.')

    x_min = int((nx - numPix) / 2)
    y_min = int((ny - numPix) / 2)
    x_max = nx - x_min
    y_max = ny - y_min
    resized = image[x_min:x_max, y_min:y_max]
    # return copy.deepcopy(resized)
    return resized  # No need to copy, since JAX returns a new DeviceArray

def re_size(image, factor=1):
    """
    re-sizes image with nx x ny to nx/factor x ny/factor
    :param image: 2d image with shape (nx,ny)
    :param factor: integer >=1
    :return:
    """
    if factor < 1:
        raise ValueError('scaling factor in re-sizing %s < 1' %factor)
    elif factor == 1:
        return image
    f = int(factor)
    nx, ny = np.shape(image)
    if int(nx/f) == nx/f and int(ny/f) == ny/f:
        # small = image.reshape([int(nx/f), f, int(ny/f), f]).mean(3).mean(1)
        image_nchw = image[None, None, :, :]
        small_nchw = functional.average_pool_2d(image_nchw, size=f, padding=ConvPadding.SAME)
        small = jnp.squeeze(small_nchw)
        return small
    else:
        raise ValueError("scaling with factor %s is not possible with grid size %s, %s" %(f, nx, ny))

def re_size_array(x_in, y_in, input_values, x_out, y_out):
    """
    resizes 2d array (i.e. image) to new coordinates. So far only works with square output aligned with coordinate axis.
    :param x_in:
    :param y_in:
    :param input_values:
    :param x_out:
    :param y_out:
    :return:
    """
    interp_2d = RectBivariateSpline(x_in, y_in, z=input_values, kx=1, ky=1, s=0)
    out_values = interp_2d(x_out, y_out)
    return out_values



# Estimation of model parameters storage and manipulation
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


from copy import deepcopy
import numpy as np
import jax.numpy as jnp
from jax import lax, jit
from functools import partial

from herculens.MassModel.Profiles import pixelated as pixelated_lens
from herculens.MassModel.Profiles import (epl, sie, sis, nie, shear, point_mass, 
                                          gaussian_potential, multipole)
from herculens.LightModel.Profiles import pixelated as pixelated_light
from herculens.LightModel.Profiles import gaussian, sersic, uniform, shapelets
from herculens.MassModel.mass_model_base import SUPPORTED_MODELS as MASS_MODELS
from herculens.LightModel.light_model_base import SUPPORTED_MODELS as LIGHT_MODELS
from herculens.Util import model_util


__all__ = ['Parameters']


class Parameters(object):
    """Class that manages parameters in JAX / auto-differentiable framework.
    Currently, it handles:
    - conversions from the differentiable parameter vector to user-friendly dictionnaries (args-to-kwargs, kwargs-to-args)
    - uniform and gaussian priors for subsets of parameters
    - log-prior values, that are meant to be added to the full loss function
    - nice LaTeX format for parameter names

    Example syntax for kwargs_joint:
    'lens_with_lens_light' : [
        [
            (lens_light_index1, lens_index1), [(lens_light_param1, lens_param1), (lens_light_param2, lens_param2), ...]
        ],
        [
            (lens_light_index2, lens_index2), [(lens_light_param3, lens_param3), (lens_light_param4, lens_param4), ...]
        ],
        ...
    ]

    Alternatively, if the parameters share the same name:
    'lens_with_lens_light' : [
        [
            (lens_light_index1, lens_index1), [param1, param2, ...]
        ],
        [
            (lens_light_index2, lens_index2), [param3, param4, ...]
        ],
        ...
    ]
    """

    _unif_prior_penalty = 1e10

    def __init__(self, lens_image, kwargs_init, kwargs_fixed, 
                 kwargs_prior=None, kwargs_joint=None):
        self._image = lens_image
        self._kwargs_init  = kwargs_init
        self._kwargs_fixed = kwargs_fixed
        num_lens_profiles = max(1, len(self._image.MassModel.profile_type_list))
        num_source_profiles = max(1, len(self._image.SourceModel.profile_type_list))
        num_lens_light_profiles = max(1, len(self._image.LensLightModel.profile_type_list))
        kwargs_prior_tmp = {
            'kwargs_lens': [{} for _ in range(num_lens_profiles)],
            'kwargs_source': [{} for _ in range(num_source_profiles)],
            'kwargs_lens_light': [{} for _ in range(num_lens_light_profiles)],
        }
        if kwargs_prior is not None:
            kwargs_prior_tmp.update(kwargs_prior)
        self._kwargs_prior = kwargs_prior_tmp
        kwargs_joint_tmp = {
            'lens_with_lens': [],
            'source_with_source': [],
            'lens_light_with_lens_light': [],
            'lens_with_lens_light': [],
        }
        if kwargs_joint is not None:
            kwargs_joint_tmp.update(kwargs_joint)
        self._kwargs_joint = kwargs_joint_tmp
        self._update_arrays()

        # TODO: write function that checks that no fields are missing
        # and fill those with default values if needed

    @property
    def optimized(self):
        return hasattr(self, '_kwargs_map')

    @property
    def num_parameters(self):
        return self._num_params

    @property
    def prior_types(self):
        return self._prior_types

    @property
    def bounds(self):
        return self._lowers, self._uppers

    def draw_prior_samples(self, num_samples, seed=None):
        """for a parameter that has no prior, returns its current value"""
        if seed is not None:
            np.random.seed(seed)
        samples = []
        for n in range(num_samples):
            param_values = []
            for i in range(self.num_parameters):
                if self._prior_types[i] == 'gaussian':
                    param_value = self._means[i] + self._widths[i] * np.random.randn()
                elif self._prior_types[i] == 'uniform':
                    param_value = np.random.uniform(low=self._lowers[i], high=self._uppers[i])
                else:
                    param_value = self.current_values()[i]
                param_values.append(param_value)
            samples.append(param_values)
        return np.array(samples)

    @property
    def names(self):
        if not hasattr(self, '_names'):
            self._names = self._set_names('mass_model_list', 'kwargs_lens')
            self._names += self._set_names('source_model_list', 'kwargs_source')
            self._names += self._set_names('lens_light_model_list', 'kwargs_lens_light')
        return self._names

    @property
    def symbols(self):
        if not hasattr(self, '_symbols'):
            self._symbols = self._name2latex(self.names)
        return self._symbols

    @property
    def kwargs_model(self):
        # TODO: intermediate step, this might be suppressed in the future
        self._kwargs_model = dict(mass_model_list=self._image.MassModel.profile_type_list,
                                  source_model_list=self._image.SourceModel.profile_type_list,
                                  lens_light_model_list=self._image.LensLightModel.profile_type_list)
        return self._kwargs_model

    def initial_values(self, as_kwargs=False, copy=False):
        if as_kwargs:
            return deepcopy(self._kwargs_init) if copy else self._kwargs_init
        else:
            return deepcopy(self._init_values) if copy else self._init_values

    def current_values(self, as_kwargs=False, restart=False, copy=False):
        if restart is True or not self.optimized:
            return self.initial_values(as_kwargs=as_kwargs, copy=copy)
        return self.best_fit_values(as_kwargs=as_kwargs, copy=copy)

    def best_fit_values(self, as_kwargs=False, copy=False):
        """Maximum-a-postriori estimate"""
        if as_kwargs:
            return deepcopy(self._kwargs_map) if copy else self._kwargs_map
        else:
            return deepcopy(self._map_values) if copy else self._map_values

    def samples(self, as_kwargs=False, group_by_param=False):
        if not hasattr(self, '_samples'):
            return None
        if as_kwargs:
            # TODO: save in cache + create separate class ParametersSamples to handle this in a cleaner way
            num_samples = self._samples.shape[0]
            kwargs_samples = [self.args2kwargs(self._samples[i, :]) for i in range(num_samples)]
            if group_by_param:
                kwargs_samples = self._group_kwargs_samples_by_param(kwargs_samples)
            return kwargs_samples
        return self._samples

    @property
    def covariance_matrix(self):
        if hasattr(self, '_cov_matrix'):
            return self._cov_matrix
        elif hasattr(self, '_samples'):
            raise NotImplementedError("Estimate covariance from samples is not yet supported.")
        else:
            return None

    def set_best_fit(self, args):
        self._map_values = args
        self._kwargs_map = self.args2kwargs(self._map_values)
    
    def set_posterior_samples(self, samples, losses):
        min_loss_idx = np.argmin(losses)
        self._map_values = samples[min_loss_idx, :]
        self._kwargs_map = self.args2kwargs(self._map_values)
        self._samples = samples

    def set_posterior_covariance(self, cov_matrix, num_samples=1000, seed=None):
        self._cov_matrix = cov_matrix
        self._samples = model_util.draw_samples_from_covariance(self.best_fit_values(),
                                                                self._cov_matrix,
                                                                num_samples=num_samples, 
                                                                seed=seed)

    def update_fixed(self, kwargs_fixed, kwargs_prior=None):
        # TODO: fill current and init values with values that were previously fixed, if needed
        self._set_params_update_fixed(kwargs_fixed, 'mass_model_list', 'kwargs_lens')
        self._set_params_update_fixed(kwargs_fixed, 'source_model_list', 'kwargs_source')
        self._set_params_update_fixed(kwargs_fixed, 'lens_light_model_list', 'kwargs_lens_light')

        # update fixed settings and everything that depends on it
        self._kwargs_fixed.update(kwargs_fixed)
        if kwargs_prior is not None:
            self._kwargs_prior = kwargs_prior
        self._update_arrays()

    # @partial(jit, static_argnums=(0,))
    def args2kwargs(self, args):
        i = 0
        args = jnp.atleast_1d(args)
        kwargs_lens, i = self._get_params(args, i, 'mass_model_list', 'kwargs_lens')
        kwargs_source, i = self._get_params(args, i, 'source_model_list', 'kwargs_source')
        kwargs_lens_light, i = self._get_params(args, i, 'lens_light_model_list', 'kwargs_lens_light')
        # apply joint param rules
        kwargs_lens = self._join_params(kwargs_lens, kwargs_lens, self._kwargs_joint['lens_with_lens'])
        kwargs_source = self._join_params(kwargs_source, kwargs_source, self._kwargs_joint['source_with_source'])
        kwargs_lens_light = self._join_params(kwargs_lens_light, kwargs_lens_light, self._kwargs_joint['lens_light_with_lens_light'])
        kwargs_lens = self._join_params(kwargs_lens_light, kwargs_lens, self._kwargs_joint['lens_with_lens_light'])
        # wrap-up
        kwargs = {'kwargs_lens': kwargs_lens, 'kwargs_source': kwargs_source, 'kwargs_lens_light': kwargs_lens_light}
        return kwargs

    # @partial(jit, static_argnums=(0,))
    def kwargs2args(self, kwargs):
        args = self._set_params(kwargs, 'mass_model_list', 'kwargs_lens')
        args += self._set_params(kwargs, 'source_model_list', 'kwargs_source')
        args += self._set_params(kwargs, 'lens_light_model_list', 'kwargs_lens_light')
        return jnp.array(args)

    def kwargs2args_prior(self, kwargs_prior):
        types_m, lowers_m, uppers_m, means_m, widths_m = self._set_params_prior(kwargs_prior, 'mass_model_list', 'kwargs_lens')
        types_s, lowers_s, uppers_s, means_s, widths_s = self._set_params_prior(kwargs_prior, 'source_model_list', 'kwargs_source')
        types_l, lowers_l, uppers_l, means_l, widths_l = self._set_params_prior(kwargs_prior, 'lens_light_model_list', 'kwargs_lens_light')
        types =  types_m  + types_s  + types_l
        lowers = lowers_m + lowers_s + lowers_l
        uppers = uppers_m + uppers_s + uppers_l
        means  = means_m  + means_s  + means_l
        widths = widths_m + widths_s + widths_l
        return types, np.array(lowers), np.array(uppers), np.array(means), np.array(widths)

    def log_prior(self, args):
        logP = 0
        for i in range(self.num_parameters):
            gaussian_prior = self._prior_types[i] == 'gaussian'
            uniform_prior  = self._prior_types[i] == 'uniform'
            logP += lax.cond(gaussian_prior, lambda _: - 0.5 * ((args[i] - self._means[i]) / self._widths[i]) ** 2, lambda _: 0., operand=None)
            logP += lax.cond(uniform_prior, lambda _: lax.cond(args[i] < self._lowers[i], lambda _: - self._unif_prior_penalty, lambda _: 0., operand=None), lambda _: 0., operand=None)
            logP += lax.cond(uniform_prior, lambda _: lax.cond(args[i] > self._uppers[i], lambda _: - self._unif_prior_penalty, lambda _: 0., operand=None), lambda _: 0., operand=None)
        return logP

    def log_prior_gaussian(self, args):
        logP = 0
        for i in range(self.num_parameters):
            gaussian_prior = self._prior_types[i] == 'gaussian'
            logP += lax.cond(gaussian_prior, lambda _: - 0.5 * ((args[i] - self._means[i]) / self._widths[i]) ** 2, lambda _: 0., operand=None)
        return logP

    def log_prior_uniform(self, args):
        logP = 0.
        for i in range(self.num_parameters):
            uniform_prior  = self._prior_types[i] == 'uniform'
            # logP += lax.cond(uniform_prior, lambda _: lax.cond(args[i] < self._lowers[i], lambda _: - self._unif_prior_penalty, lambda _: 0., operand=None), lambda _: 0., operand=None)
            # logP += lax.cond(uniform_prior, lambda _: lax.cond(args[i] > self._uppers[i], lambda _: - self._unif_prior_penalty, lambda _: 0., operand=None), lambda _: 0., operand=None)
            logP += lax.cond(uniform_prior, lambda _: - (args[i] - jnp.clip(args[i], a_min=self._lowers[i], a_max=self._uppers[i]))**2, lambda _: 0., operand=None)
        return logP

    def apply_bounds(self, args):
        return jnp.clip(args, a_min=self._lowers, a_max=self._uppers)

    def log_prior_nojit(self, args):
        logP = 0
        for i in range(self.num_parameters):
            if self._prior_types[i] == 'gaussian':
                logP += - 0.5 * ((args[i] - self._means[i]) / self._widths[i]) ** 2
            elif self._prior_types[i] == 'uniform' and not (self._lowers[i] <= args[i] <= self._uppers[i]):
                logP += - self._unif_prior_penalty
        return logP

    @staticmethod
    def get_class_for_model(kwargs_key, model):
        # TODO: move outside of this class
        profile_class = None
        if kwargs_key in ['kwargs_source', 'kwargs_lens_light']:
            if model not in LIGHT_MODELS:
                raise ValueError(f"'{model}' is not supported.")
            if model == 'GAUSSIAN':
                profile_class = gaussian.Gaussian
            if model == 'GAUSSIAN_ELLIPSE':
                profile_class = gaussian.GaussianEllipse
            elif model == 'SERSIC':
                profile_class = sersic.Sersic
            elif model == 'SERSIC_ELLIPSE':
                profile_class = sersic.SersicElliptic
            elif model == 'CORE_SERSIC':
                profile_class = sersic.CoreSersic
            elif model == 'UNIFORM':
                profile_class = uniform.Uniform
            elif model == 'PIXELATED':
                profile_class = pixelated_light.Pixelated
            elif model == 'SHAPELETS':
                profile_class = shapelets.Shapelets
        elif kwargs_key == 'kwargs_lens':
            if model not in MASS_MODELS:
                raise ValueError("'{model}' is not supported.")
            elif model == 'GAUSSIAN':
                profile_class = gaussian_potential.Gaussian
            elif model == 'EPL':
                profile_class = epl.EPL
            elif model == 'SIE':
                profile_class = sie.SIE
            elif model == 'SIS':
                profile_class = sis.SIS
            elif model == 'NIE':
                profile_class = nie.NIE
            elif model == 'POINT_MASS':
                profile_class = point_mass.PointMass
            elif model == 'SHEAR':
                profile_class = shear.Shear
            elif model == 'SHEAR_GAMMA_PSI':
                profile_class = shear.ShearGammaPsi
            elif model == 'MULTIPOLE':
                profile_class = multipole.Multipole
            elif model == 'PIXELATED':
                profile_class = pixelated_lens.PixelatedPotential
            elif model == 'PIXELATED_DIRAC':
                profile_class = pixelated_lens.PixelatedPotentialDirac
        if profile_class is None:
            raise ValueError(f"Could not find the model class for '{model}'")
        return profile_class

    @staticmethod
    def get_param_names_for_model(kwargs_key, model):
        return Parameters.get_class_for_model(kwargs_key, model).param_names

    def _update_arrays(self):
        self._kwargs_fixed = self._update_fixed_with_joint(self._kwargs_fixed, self._kwargs_joint)
        self._prior_types, self._lowers, self._uppers, self._means, self._widths \
            = self.kwargs2args_prior(self._kwargs_prior)
        self._init_values = self.kwargs2args(self._kwargs_init)
        self._kwargs_init = self.args2kwargs(self._init_values)  # for updating missing fields
        self._num_params = len(self._init_values)
        if self.optimized:
            self._map_values = self.kwargs2args(self._kwargs_map)
        if hasattr(self, '_names'):
            delattr(self, '_names')
        if hasattr(self, '_symbols'):
            delattr(self, '_symbols')
        if hasattr(self, '_samples'):
            delattr(self, '_samples')
        if hasattr(self, '_kwargs_samples'):
            delattr(self, '_kwargs_samples')

    def _update_fixed_with_joint(self, kwargs_fixed, kwargs_joint):
        kwargs_fixed = self._update_fixed_with_joint_one(kwargs_fixed, kwargs_joint, 'kwargs_lens', 'lens_with_lens')
        kwargs_fixed = self._update_fixed_with_joint_one(kwargs_fixed, kwargs_joint, 'kwargs_source', 'source_with_source')
        kwargs_fixed = self._update_fixed_with_joint_one(kwargs_fixed, kwargs_joint, 'kwargs_lens_light', 'lens_light_with_lens_light')
        kwargs_fixed = self._update_fixed_with_joint_one(kwargs_fixed, kwargs_joint, 'kwargs_lens', 'lens_with_lens_light')
        return kwargs_fixed

    @staticmethod
    def _update_fixed_with_joint_one(kwargs_fixed, kwargs_joint, kwargs_key, joint_key):
        kwargs_fixed_updt = deepcopy(kwargs_fixed)
        joint_setting_list = kwargs_joint[joint_key]
        for setting in joint_setting_list:
            (i_1, k_2), param_list = setting
            for param_names in param_list:
                if isinstance(param_names, (tuple, list)) and len(param_names) > 1:
                    param_name_1, param_name_2 = param_names
                else:
                    param_name_1 = param_name_2 = param_names
                kwargs_fixed_updt[kwargs_key][k_2][param_name_2] = 0
        return kwargs_fixed_updt

    @staticmethod
    def _join_params(kwargs_list_1, kwargs_list_2, joint_setting_list):
        for setting in joint_setting_list:
            (i_1, k_2), param_list = setting
            for param_names in param_list:
                if isinstance(param_names, (tuple, list)) and len(param_names) > 1:
                    param_name_1, param_name_2 = param_names
                else:
                    param_name_1 = param_name_2 = param_names
                kwargs_list_2[k_2][param_name_2] = kwargs_list_1[i_1][param_name_1]
        return kwargs_list_2

    def _get_params(self, args, i, kwargs_model_key, kwargs_key):
        kwargs_list = []
        for k, model in enumerate(self.kwargs_model[kwargs_model_key]):
            kwargs = {}
            kwargs_fixed_k = self._kwargs_fixed[kwargs_key][k]
            param_names = self.get_param_names_for_model(kwargs_key, model)
            for name in param_names:
                if not name in kwargs_fixed_k:
                    if model == 'PIXELATED':
                        if kwargs_key == 'kwargs_lens':
                            n_pix_x, n_pix_y = self._image.MassModel.pixelated_shape
                        elif kwargs_key == 'kwargs_source':
                            n_pix_x, n_pix_y = self._image.SourceModel.pixelated_shape
                        elif kwargs_key == 'kwargs_lens_light':
                            n_pix_x, n_pix_y = self._image.LensLightModel.pixelated_shape
                        num_param = int(n_pix_x * n_pix_y)
                        kwargs['pixels'] = args[i:i + num_param].reshape(n_pix_x, n_pix_y)
                    elif model == 'SHAPELETS' and name == 'amps':
                        if kwargs_key == 'kwargs_source':
                            num_param = self._image.SourceModel.num_amplitudes_list[k]
                        elif kwargs_key == 'kwargs_lens_light':
                            num_param = self._image.LensLightModel.num_amplitudes_list[k]
                        else:
                            raise ValueError("Basis functions can only be in the source or lens light.")
                        kwargs['amps'] = args[i:i + num_param]
                    else:
                        num_param = 1
                        kwargs[name] = args[i]
                    i += num_param
                else:
                    kwargs[name] = kwargs_fixed_k[name]
            kwargs_list.append(kwargs)
        return kwargs_list, i

    def _set_params(self, kwargs, kwargs_model_key, kwargs_key):
        args = []
        for k, model in enumerate(self.kwargs_model[kwargs_model_key]):
            kwargs_profile = kwargs[kwargs_key][k]
            kwargs_fixed_k = self._kwargs_fixed[kwargs_key][k]
            param_names = self.get_param_names_for_model(kwargs_key, model)
            for name in param_names:
                if not name in kwargs_fixed_k:
                    if model == 'PIXELATED':
                        pixels = kwargs_profile['pixels']
                        if kwargs_key == 'kwargs_lens':
                            n_pix_x, n_pix_y = self._image.MassModel.pixelated_shape
                        elif kwargs_key == 'kwargs_source':
                            n_pix_x, n_pix_y = self._image.SourceModel.pixelated_shape
                        elif kwargs_key == 'kwargs_lens_light':
                            n_pix_x, n_pix_y = self._image.LensLightModel.pixelated_shape
                        if isinstance(pixels, (int, float)):
                            pixels = pixels * np.ones((n_pix_x, n_pix_y))
                        elif pixels.shape != (n_pix_x, n_pix_y):
                            raise ValueError("Pixelated array is inconsistent with pixelated grid.")
                        args += pixels.flatten().tolist()
                    elif model == 'SHAPELETS' and name == 'amps':
                        amps = kwargs_profile['amps']
                        if kwargs_key == 'kwargs_source':
                            num_param = self._image.SourceModel.num_amplitudes_list[k]
                        elif kwargs_key == 'kwargs_lens_light':
                            num_param = self._image.LensLightModel.num_amplitudes_list[k]
                        else:
                            raise ValueError("Basis functions can only be in the source or lens light.")
                        if len(amps) != num_param:
                            raise ValueError("Number of functions' amplitudes is not the on expected.")
                        args += np.asarray(amps).flatten().tolist()
                    else:
                        args.append(kwargs_profile[name])
        return args

    def _set_params_prior(self, kwargs, kwargs_model_key, kwargs_key):
        types, lowers, uppers, means, widths = [], [], [], [], []
        for k, model in enumerate(self.kwargs_model[kwargs_model_key]):
            kwargs_profile = kwargs[kwargs_key][k]
            kwargs_fixed_k = self._kwargs_fixed[kwargs_key][k]
            param_names = self.get_param_names_for_model(kwargs_key, model)
            for name in param_names:
                if not name in kwargs_fixed_k:
                    if name not in kwargs_profile:
                        prior_type = None
                    else:
                        prior_type = kwargs_profile[name][0]

                    if model == 'PIXELATED':
                        if kwargs_key == 'kwargs_lens':
                            n_pix_x, n_pix_y = self._image.MassModel.pixelated_shape
                        elif kwargs_key == 'kwargs_source':
                            n_pix_x, n_pix_y = self._image.SourceModel.pixelated_shape
                        elif kwargs_key == 'kwargs_lens_light':
                            n_pix_x, n_pix_y = self._image.LensLightModel.pixelated_shape
                        num_param = int(n_pix_x * n_pix_y)
                    elif model == 'SHAPELETS' and name == 'amps':
                        if kwargs_key == 'kwargs_source':
                            num_param = self._image.SourceModel.num_amplitudes_list[k]
                        elif kwargs_key == 'kwargs_lens_light':
                            num_param = self._image.LensLightModel.num_amplitudes_list[k]
                    else:
                        num_param = 1
                    
                    if prior_type == 'uniform':
                        if num_param > 1:
                            types += [prior_type]*num_param
                            lowers_tmp, uppers_tmp = kwargs_profile['pixels'][1], kwargs_profile['pixels'][2]
                            # those bounds can either be whole array (values per pixel)
                            if isinstance(lowers_tmp, (np.ndarray, jnp.ndarray)):
                                lowers += lowers_tmp.flatten().tolist()
                                uppers += uppers_tmp.flatten().tolist()
                            # or they can be single numbers, in which case they are considered the same for pixel
                            elif isinstance(lowers_tmp, (int, float)):
                                lowers += [float(lowers_tmp)]*num_param
                                uppers += [float(uppers_tmp)]*num_param
                            means  += [np.nan]*num_param
                            widths += [np.nan]*num_param
                        else:
                            types.append(prior_type)
                            lowers.append(float(kwargs_profile[name][1]))
                            uppers.append(float(kwargs_profile[name][2]))
                            means.append(np.nan)
                            widths.append(np.nan)

                    elif prior_type == 'gaussian':
                        if model in ['PIXELATED', 'SHAPELETS']:
                            raise ValueError(f"'gaussian' prior for '{model}' model is not supported")
                        else:
                            types.append(prior_type)
                            lowers.append(-np.inf)
                            uppers.append(+np.inf)
                            means.append(kwargs_profile[name][1])
                            widths.append(kwargs_profile[name][2])

                    else:
                        types  += [prior_type]*num_param
                        lowers += [-np.inf]*num_param
                        uppers += [+np.inf]*num_param
                        means  += [np.nan]*num_param
                        widths += [np.nan]*num_param

        return types, lowers, uppers, means, widths

    def _set_params_update_fixed(self, kwargs_fixed, kwargs_model_key, kwargs_key):
        for k, model in enumerate(self.kwargs_model[kwargs_model_key]):
            kwargs_fixed_k_old = self._kwargs_fixed[kwargs_key][k]
            kwargs_fixed_k_new = kwargs_fixed
            param_names = self.get_param_names_for_model(kwargs_key, model)
            for name in param_names:
                if name in kwargs_fixed_k_old and name not in kwargs_fixed_k_new:
                    self._kwargs_init[kwargs_key][k][name] = deepcopy(kwargs_fixed_k_old[name])
                    if self.optimized:
                        self._kwargs_map[kwargs_key][k][name] = deepcopy(kwargs_fixed_k_old[name])

    @partial(jit, static_argnums=(0,))
    def _group_kwargs_samples_by_param(self, kwargs_samples):
        kwargs_samples_grouped = {}
        num_samples = len(kwargs_samples)
        for kwargs_model_key, kwargs_model in kwargs_samples[0].items():
            kwargs_samples_grouped[kwargs_model_key] = []
            for k in range(len(kwargs_model)):
                kwargs_samples_grouped[kwargs_model_key].append({})
                for param_name in kwargs_samples[0][kwargs_model_key][k].keys():
                    samples = []
                    for i in range(num_samples):
                        param_values_i = kwargs_samples[i][kwargs_model_key][k][param_name]
                        samples.append(param_values_i)
                    kwargs_samples_grouped[kwargs_model_key][k][param_name] = samples
        return kwargs_samples_grouped

    def _set_names(self, kwargs_model_key, kwargs_key):
        names = []
        short_id = kwargs_key.replace('kwargs_', '')
        for k, model in enumerate(self.kwargs_model[kwargs_model_key]):
            kwargs_fixed_k = self._kwargs_fixed[kwargs_key][k]
            param_names = self.get_param_names_for_model(kwargs_key, model)
            for name in param_names:
                if not name in kwargs_fixed_k:
                    if model == 'PIXELATED':
                        if kwargs_key == 'kwargs_lens':
                            n_pix_x, n_pix_y = self._image.MassModel.pixelated_shape
                            num_param = int(n_pix_x * n_pix_y)
                            names_k = [f"psi_{i}" for i in range(num_param)]  # 'psi' for deflector
                        elif kwargs_key == 'kwargs_source':
                            n_pix_x, n_pix_y = self._image.SourceModel.pixelated_shape
                            num_param = int(n_pix_x * n_pix_y)
                            names_k = [f"s_{i}" for i in range(num_param)]  # 's' for source
                        elif kwargs_key == 'kwargs_lens_light':
                            n_pix_x, n_pix_y = self._image.LensLightModel.pixelated_shape
                            num_param = int(n_pix_x * n_pix_y)
                            names_k = [f"l_{i}" for i in range(num_param)]  # 'l' for potential
                    elif model == 'SHAPELETS' and name == 'amps':
                        if kwargs_key == 'kwargs_source':
                            num_param = self._image.SourceModel.num_amplitudes_list[k]
                        elif kwargs_key == 'kwargs_lens':
                            num_param = self._image.LensLightModel.num_amplitudes_list[k]
                        names_k = [f"amp_{i}" for i in range(num_param)]
                    else:
                        names_k = [name]
                    names += [f"{n}-{short_id}-{k}" for n in names_k]  # assign a unique identifier
        return names

    @staticmethod
    def name2latex(name_raw):
        # TODO: move outside of this class
        name, model_type, profile_idx = name_raw.split('-')   # encapsulate this line in a well-named method

        # pixelated models
        if name[:2] == 's_':  
            latex = r"$s_{" + r"{}".format(int(name[2:])) + r"}$"
        elif name[:2] == 'l_':  
            latex = r"$l_{" + r"{}".format(int(name[2:])) + r"}$"
        elif name[:4] == 'psi_':  
            latex = r"$\psi_{" + r"{}".format(int(name[4:])) + r"}$"

        # analytical models
        elif name == 'theta_E':
            latex = r"$\theta_{\rm E}$"
        elif name == 'gamma':
            latex = r"$\gamma'$"
        elif name == 'gamma_ext':
            latex = r"$\gamma_{\rm ext}$"
        elif name == 'psi_ext':
            latex = r"$\psi_{\rm ext}$"
        elif name == 'gamma1':
            latex = r"$\gamma_{\rm 1, ext}$"
        elif name == 'gamma2':
            latex = r"$\gamma_{\rm 2, ext}$"
        elif name == 'amp':
            latex = r"$A$"
        elif name == 'R_sersic':
            latex = r"$R_{\rm Sersic}$"
        elif name == 'n_sersic':
            latex = r"$n_{\rm Sersic}$"
        elif name == 'e1':
            latex = r"$e_1$"
        elif name == 'e2':
            latex = r"$e_2$"
        elif name == 'center_x':
            latex = r"$x_0$"
        elif name == 'center_y':
            latex = r"$y_0$"
        elif name == 'ra_0':
            latex = r"${\rm RA}_0$"
        elif name == 'dec_0':
            latex = r"${\rm Dec}_0$"
        elif name == 'm':
            latex = r"$m$"
        elif name == 'a_m':
            latex = r"$a_m$"
        elif name == 'phi_m':
            latex = r"$\phi_m$"
        else:
            raise ValueError("latex symbol for variable '{}' is unknown".format(name))
        return latex

    def _name2latex(self, names):
        latexs = []
        for name in names:
            latexs.append(self.name2latex(name))
        return latexs



# Defines a general fully differentiable probability function for inference purposes
# 
# Copyright (c) 2022, herculens developers and contributors

__author__ = 'aymgal'


from functools import partial
from jax import jit


__all__ = ['Inference']


class Inference(object):

    """Abstract class that defines wraps the loss function, and defines first and second order derivatives.
    :param loss_class: herculens.Inference.loss.Loss instance
    :param param_class: herculens.Parameters.parameters.Parameters instance
    """

    def __init__(self, loss_class):
        self._loss = loss_class
        self.kinetic_fn = None  # for numpyro HMC, will default to Euclidean kinetic energy

    @partial(jit, static_argnums=(0,))
    def log_probability(self, args):
        """unnormalized log-posterior as the negative loss, typically for Bayesian inference using MCMC"""
        return - self._loss(args)

    @partial(jit, static_argnums=(0,))
    def potential_fn(self, args):
        """alias for negative log-likelihood, typically for Bayesian inference using HMC"""
        return self._loss(args)
        


# Estimation of the parameter covariance matrix via the Fisher information
# 
# Copyright (c) 2022, herculens developers and contributors

__author__ = 'aymgal'


import numpy as np
from herculens.Util import model_util


__all__ = ['FisherCovariance']


class FisherCovariance(object):

    def __init__(self, parameter_class, differentiable_class):
        self._param = parameter_class  # TODO: this should hold the indices for splitting matrices
        self._diff = differentiable_class

    @property
    def fisher_matrix(self):
        if not hasattr(self, '_fim'):
            raise ValueError("Call first compute_fisher_information().")
        return self._fim

    @property
    def covariance_matrix(self):
        if not hasattr(self, '_cov'):
            self._cov = self.fisher2covar(self.fisher_matrix, inversion='full')
        return self._cov

    def model_covariance(self, lens_image, num_mc_samples=10000, seed=None,
                         return_cross_covariance=False):
        samples = self.draw_samples(num_samples=num_mc_samples, seed=seed)
        return model_util.estimate_model_covariance(lens_image, self._param, samples,
                                                    return_cross_covariance=return_cross_covariance)

    def draw_samples(self, num_samples=10000, seed=None):
        return model_util.draw_samples_from_covariance(self._param.best_fit_values(),
                                                       self.covariance_matrix,
                                                       num_samples=num_samples, seed=seed)

    def get_kwargs_sigma(self):
        sigma_values = np.sqrt(np.abs(np.diag(self.covariance_matrix)))
        return self._param.args2kwargs(sigma_values)

    def compute_fisher_information(self, recompute=False):
        if hasattr(self, '_fim') and not recompute:
            return  # nothing to do
        best_fit_values = self._param.best_fit_values()
        self._fim = self._diff.hessian(best_fit_values).block_until_ready()
        self._fim = np.array(self._fim)
        if hasattr(self, '_cov'):
            delattr(self, '_cov')

    def fisher2covar(self, fisher_matrix, inversion='full'):
        if inversion == 'full':
            return np.linalg.inv(fisher_matrix)
        elif inversion == 'diag':
            return 1. / np.diag(fisher_matrix)
        else:
            raise ValueError("Only 'full' and 'diag' options are supported for inverting the FIM.")

    @staticmethod
    def split_matrix(matrix, num_before, num_after):
        interior = matrix[num_before:-num_after, num_before:-num_after]

        block_UL = matrix[:num_before, :num_before]
        block_UR = matrix[:num_before, -num_after:]
        block_LR = matrix[-num_after:, -num_after:]
        block_LL = matrix[-num_after:, :num_before]
        exterior = np.block([[block_UL, block_UR], 
                            [block_LL, block_LR]])

        return interior, exterior



# Handles different method to optimize a loss function
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal', 'austinpeel'


import time
from copy import deepcopy
from functools import partial
from jax import jit
import jax.numpy as jnp
import optax

from herculens.Inference.Optimization.base_optim import BaseOptimizer


__all__ = ['OptaxOptimizer']


class OptaxOptimizer(BaseOptimizer):
    """Wrapper to optax's gradient descent optimizers"""

    def __init__(self, loss, projection_fn=None, **kwargs):
        super().__init__(loss, **kwargs)
        self._projection_fn = projection_fn

    # TODO: output the loss value next the progressbar

    def run(self, init_params, algorithm='adabelief', max_iterations=100, min_iterations=None,
            init_learning_rate=1e-2, schedule_learning_rate=True, 
            stop_at_loss_increase=False, progress_bar=True, return_param_history=False):
        if min_iterations is None:
            min_iterations = max_iterations
        if schedule_learning_rate is True:
            # Exponential decay of the learning rate
            scheduler = optax.exponential_decay(
                init_value=init_learning_rate, 
                decay_rate=0.99, # NOTE: this has never been fine-tuned (taken from optax examples)
                transition_steps=max_iterations)

            if algorithm.lower() == 'adabelief':
                scale_algo = optax.scale_by_belief()
            elif algorithm.lower() == 'radam':
                scale_algo = optax.scale_by_radam()
            elif algorithm.lower() == 'adam':
                scale_algo = optax.scale_by_adam()
            else:
                raise ValueError(f"Optax algorithm '{algorithm}' is not supported")

            # Combining gradient transforms using `optax.chain`
            optim = optax.chain(
                #optax.clip_by_global_norm(1.0),  # clip by the gradient by the global norm # TODO: what is this used for?
                scale_algo,  # use the updates from the chosen optimizer
                optax.scale_by_schedule(scheduler),  # Use the learning rate from the scheduler
                optax.scale(-1.)  # because gradient *descent*
            )
        else:
            if algorithm.lower() == 'adabelief':
                optim = optax.adabelief(init_learning_rate)
            elif algorithm.lower() == 'radam':
                optim = optax.radam(init_learning_rate)
            elif algorithm.lower() == 'adam':
                optim = optax.adam(init_learning_rate)
            else:
                raise ValueError(f"Optax algorithm '{algorithm}' is not supported")

        # Initialise optimizer state
        #params = self._param.current_values(as_kwargs=False, restart=restart_from_init, copy=True)
        params = deepcopy(init_params)
        opt_state = optim.init(params)
        prev_params, prev_loss_val = params, 1e10

        @jit
        def gd_step(params, opt_state):
            loss_val, grads = self.function_optim_with_grad(params)
            updates, opt_state = optim.update(grads, opt_state, params)
            params = optax.apply_updates(params, updates)
            params = self.apply_projections(params)
            return params, opt_state, loss_val

        # Gradient descent loop
        param_history = []
        loss_history = []
        start_time = time.time()
        for i in self._for_loop(range(max_iterations), progress_bar, 
                                total=max_iterations, 
                                desc=f"optax.{algorithm}"):
            params, opt_state, loss_val = gd_step(params, opt_state)
            if stop_at_loss_increase and i > min_iterations and loss_val > prev_loss_val:
                params, loss_val = prev_params, prev_loss_val
                break
            else:
                loss_history.append(loss_val)
                prev_params, prev_loss_val = params, loss_val
            if return_param_history is True:
                param_history.append(params)
        runtime = time.time() - start_time
        best_fit = params
        logL_best_fit = - self.loss.function(best_fit)
        extra_fields = {'loss_history': jnp.array(loss_history)}  # TODO: use optax.second_order module to compute diagonal of Hessian?
        if return_param_history is True:
            extra_fields['param_history'] = param_history
        return best_fit, logL_best_fit, extra_fields, runtime

    @partial(jit, static_argnums=(0,))
    def apply_projections(self, params):
        if self._projection_fn is not None:
            params = self._projection_fn(params)
        return params
